
import numpy as np

from .external.skimage.measure import find_contours, points_in_poly
import scipy.interpolate as spint

from .kde_methods import get_bad_vals


def find_contours_level(density, x, y, level, closed=False):
    """Find iso-valued density contours for a given level value

    Parameters
    ----------
    density: 2d ndarray of shape (M, N)
        Kernel density estimate (KDE) for which to compute the contours
    x: 2d ndarray of shape (M, N) or 1d ndarray of size M
        X-values corresponding to `density`
    y: 2d ndarray of shape (M, N) or 1d ndarray of size M
        Y-values corresponding to `density`
    level: float between 0 and 1
        Value along which to find contours in `density` relative
        to its maximum
    closed: bool
        Whether to close contours at the KDE support boundaries

    Returns
    -------
    contours: list of ndarrays of shape (P, 2)
        Contours found for the given level value

    See Also
    --------
    skimage.measure.find_contours: Contour finding algorithm used
    """
    if level >= 1 or level <= 0:
        raise ValueError("`level` must be in (0,1), got '{}'!".format(level))
    # level relative to maximum
    level = level * density.max()
    # xy coordinates
    if len(x.shape) == 2:
        assert np.all(x[:, 0] == x[:, 1])
        x = x[:, 0]
    if len(y.shape) == 2:
        assert np.all(y[0, :] == y[1, :])
        y = y[0, :]
    if closed:
        # find closed contours
        density = np.pad(density, ((1, 1), (1, 1)), mode="constant")
        offset = 1
    else:
        # leave contours open at kde boundary
        offset = 0

    conts_idx = find_contours(density, level)
    conts_xy = []

    for cc in conts_idx:
        cx = np.interp(x=cc[:, 0]-offset,
                       xp=range(x.size),
                       fp=x)
        cy = np.interp(x=cc[:, 1]-offset,
                       xp=range(y.size),
                       fp=y)
        conts_xy.append(np.stack((cx, cy), axis=1))

    return conts_xy


def get_quantile_levels(density, x, y, xp, yp, q, normalize=True):
    """Compute density levels for given quantiles by interpolation

    For a given 2D density, compute the density levels at which
    the resulting contours contain the fraction `1-q` of all
    data points. E.g. for a measurement of 1000 events, all
    contours at the level corresponding to a quantile of
    `q=0.95` (95th percentile) contain 50 events (5%).

    Parameters
    ----------
    density: 2d ndarray of shape (M, N)
        Kernel density estimate for which to compute the contours
    x: 2d ndarray of shape (M, N) or 1d ndarray of size M
        X-values corresponding to `density`
    y: 2d ndarray of shape (M, N) or 1d ndarray of size M
        Y-values corresponding to `density`
    xp: 1d ndarray of size D
        Event x-data from which to compute the quantile
    yp: 1d ndarray of size D
        Event y-data from which to compute the quantile
    q: array_like or float between 0 and 1
        Quantile along which to find contours in `density` relative
        to its maximum
    normalize: bool
        Whether output levels should be normalized to the maximum
        of `density`

    Returns
    -------
    level: np.ndarray or float
        Contours level(s) corresponding to the given quantile

    Notes
    -----
    NaN-values events in `xp` and `yp` are ignored.
    """
    # xy coordinates
    if len(x.shape) == 2:
        assert np.all(x[:, 0] == x[:, 1])
        x = x[:, 0]
    if len(y.shape) == 2:
        assert np.all(y[0, :] == y[1, :])
        y = y[0, :]

    # remove bad events
    bad = get_bad_vals(xp, yp)
    xp = xp[~bad]
    yp = yp[~bad]

    # Normalize interpolation data such that the spacing for
    # x and y is about the same during interpolation.
    x_norm = x.max()
    x = x / x_norm
    xp = xp / x_norm

    y_norm = y.max()
    y = y / y_norm
    yp = yp / y_norm

    # Perform interpolation
    dp = spint.interpn((x, y), density,
                       (xp, yp),
                       method='linear',
                       bounds_error=False,
                       fill_value=0)

    if normalize:
        dp /= density.max()

    if not np.isscalar(q):
        q = np.array(q)
    plev = np.nanpercentile(dp, q=q*100)
    return plev


def _find_quantile_level(density, x, y, xp, yp, quantile, acc=.01,
                         ret_err=False):
    """Find density level for a given data quantile by iteration

    Parameters
    ----------
    density: 2d ndarray of shape (M, N)
        Kernel density estimate for which to compute the contours
    x: 2d ndarray of shape (M, N) or 1d ndarray of size M
        X-values corresponding to `density`
    y: 2d ndarray of shape (M, N) or 1d ndarray of size M
        Y-values corresponding to `density`
    xp: 1d ndarray of size D
        Event x-data from which to compute the quantile
    yp: 1d ndarray of size D
        Event y-data from which to compute the quantile
    quantile: float between 0 and 1
        Quantile along which to find contours in `density` relative
        to its maximum
    acc: float
        Desired absolute accuracy (stopping criterion) of the
        contours
    ret_err: bool
        If True, also return the absolute error

    Returns
    -------
    level: float
        Contours level corresponding to the given quantile

    Notes
    -----
    A much more faster method (using interpolation) is implemented in
    :func:`get_quantile_levels`.
    NaN-values events in `xp` and `yp` are ignored.

    See Also
    --------
    skimage.measure.find_contours: Contour finding algorithm
    """
    if quantile >= 1 or quantile <= 0:
        raise ValueError("Invalid value for `quantile`: {}".format(quantile))

    # remove bad events
    bad = get_bad_vals(xp, yp)
    xp = xp[~bad]
    yp = yp[~bad]
    points = np.concatenate((xp.reshape(-1, 1), yp.reshape(-1, 1)), axis=1)

    # initial guess
    level = quantile
    # error of current iteration
    err = 1
    # iteration factor (guarantees convergence)
    itfac = 1
    # total number of events
    nev = xp.size

    while np.abs(err) > acc:
        # compute contours
        conts = find_contours_level(density, x, y, level, closed=True)
        # compute number of points in contour
        isin = 0
        pi = np.array(points, copy=True)
        for cc in conts:
            pinc = points_in_poly(points=pi, verts=cc)
            isin += np.sum(pinc)
            # ignore these points for the other contours
            pi = pi[~pinc]
        err = quantile - (nev - isin) / nev
        level += err * itfac
        itfac *= .9

    if ret_err:
        return level, err
    else:
        return level




import os
import tempfile

import numpy as np
import pytest

import dclab

from helper_methods import example_data_dict


filter_data = """[Polygon 00000000]
X Axis = aspect
Y Axis = tilt
Name = polygon filter 0
point00000000 = 6.344607717656481e-03 7.703315881326352e-01
point00000001 = 7.771010748302133e-01 7.452006980802792e-01
point00000002 = 8.025596093384512e-01 6.806282722513089e-03
point00000003 = 6.150993521573982e-01 1.015706806282723e-03
"""


@pytest.mark.filterwarnings('ignore::dclab.polygon_filter.'
                            + 'FilterIdExistsWarning')
def test_import():
    dclab.PolygonFilter.clear_all_filters()
    ddict = example_data_dict(size=1000, keys=["aspect", "tilt"])
    ds = dclab.new_dataset(ddict)

    # save polygon data
    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)

    # Add polygon filter
    pf = dclab.PolygonFilter(filename=tf)
    ds.polygon_filter_add(pf)

    ds.apply_filter()

    assert np.sum(ds.filter.all) == 330

    dclab.PolygonFilter.import_all(tf)

    assert len(dclab.PolygonFilter.instances) == 2

    # Import multiples
    b = filter_data
    b = b.replace("Polygon 00000000", "Polygon 00000001")
    with open(tf, "a") as fd:
        fd.write(b)
    dclab.PolygonFilter.import_all(tf)

    # Import previously saved
    dclab.PolygonFilter.save_all(tf)
    dclab.PolygonFilter.import_all(tf)

    assert len(dclab.PolygonFilter.instances) == 10

    try:
        os.remove(tf)
    except OSError:
        pass


def test_invert():
    dclab.PolygonFilter.clear_all_filters()
    ddict = example_data_dict(size=1234, keys=["aspect", "tilt"])
    ds = dclab.new_dataset(ddict)
    # points of polygon filter
    points = [[np.min(ddict["aspect"]), np.min(ddict["tilt"])],
              [np.min(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.min(ddict["tilt"])],
              ]
    filt1 = dclab.PolygonFilter(axes=["aspect", "tilt"],
                                points=points,
                                inverted=False)
    ds.polygon_filter_add(filt1)
    assert [0] == ds.config["filtering"]["polygon filters"]
    n1 = np.sum(ds.filter.all)
    ds.apply_filter()
    n2 = np.sum(ds.filter.all)
    assert n1 != n2
    filt2 = dclab.PolygonFilter(axes=["aspect", "tilt"],
                                points=points,
                                inverted=True)
    ds.polygon_filter_add(filt2)
    assert [0, 1] == ds.config["filtering"]["polygon filters"]
    ds.apply_filter()
    assert np.sum(ds.filter.all) == 0, "inverted+normal filter filters all"
    dclab.PolygonFilter.clear_all_filters()


def test_invert_copy():
    dclab.PolygonFilter.clear_all_filters()
    ddict = example_data_dict(size=1234, keys=["aspect", "tilt"])
    ds = dclab.new_dataset(ddict)
    # points of polygon filter
    points = [[np.min(ddict["aspect"]), np.min(ddict["tilt"])],
              [np.min(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.min(ddict["tilt"])],
              ]
    filt1 = dclab.PolygonFilter(axes=["aspect", "tilt"],
                                points=points,
                                inverted=False)
    ds.polygon_filter_add(filt1)
    assert [0] == ds.config["filtering"]["polygon filters"]
    n1 = np.sum(ds.filter.all)
    ds.apply_filter()
    n2 = np.sum(ds.filter.all)
    assert n1 != n2
    filt2 = filt1.copy(invert=True)
    ds.polygon_filter_add(filt2)
    assert [0, 1] == ds.config["filtering"]["polygon filters"]
    ds.apply_filter()
    assert np.sum(ds.filter.all) == 0, "inverted+normal filter filters all"
    dclab.PolygonFilter.clear_all_filters()


@pytest.mark.filterwarnings('ignore::dclab.polygon_filter.'
                            + 'FilterIdExistsWarning')
def test_invert_saveload():
    dclab.PolygonFilter.clear_all_filters()
    ddict = example_data_dict(size=1234, keys=["aspect", "tilt"])
    # points of polygon filter
    points = [[np.min(ddict["aspect"]), np.min(ddict["tilt"])],
              [np.min(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.min(ddict["tilt"])],
              ]
    filt1 = dclab.PolygonFilter(axes=["aspect", "tilt"],
                                points=points,
                                inverted=True)
    name = tempfile.mktemp(prefix="test_dclab_polygon_")
    filt1.save(name)
    filt2 = dclab.PolygonFilter(filename=name)
    assert filt2 == filt1

    filt3 = dclab.PolygonFilter(axes=["aspect", "tilt"],
                                points=points,
                                inverted=False)
    try:
        os.remove(name)
    except OSError:
        pass

    name = tempfile.mktemp(prefix="test_dclab_polygon_")
    filt3.save(name)
    filt4 = dclab.PolygonFilter(filename=name)
    assert filt4 == filt3
    try:
        os.remove(name)
    except OSError:
        pass


def test_inverted_wrong():
    dclab.PolygonFilter.clear_all_filters()
    ddict = example_data_dict(size=1234, keys=["aspect", "tilt"])
    # points of polygon filter
    points = [[np.min(ddict["aspect"]), np.min(ddict["tilt"])],
              [np.min(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.max(ddict["tilt"])],
              [np.average(ddict["aspect"]), np.min(ddict["tilt"])],
              ]
    try:
        dclab.PolygonFilter(axes=["aspect", "tilt"],
                            points=points,
                            inverted=0)
    except dclab.polygon_filter.PolygonFilterError:
        pass
    else:
        raise ValueError("inverted should only be allowed to be bool")


def test_nofile_copy():
    dclab.PolygonFilter.clear_all_filters()
    a = dclab.PolygonFilter(axes=("tilt", "aspect"),
                            points=[[0, 1], [1, 1]])
    a.copy()
    dclab.PolygonFilter.clear_all_filters()


def test_remove():
    dclab.PolygonFilter.clear_all_filters()

    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)

    # Add polygon filter
    pf = dclab.PolygonFilter(filename=tf)

    dclab.PolygonFilter.remove(pf.unique_id)
    assert len(dclab.PolygonFilter.instances) == 0

    dclab.PolygonFilter.clear_all_filters()
    try:
        os.remove(tf)
    except OSError:
        pass


@pytest.mark.filterwarnings('ignore::dclab.polygon_filter.'
                            + 'FilterIdExistsWarning')
def test_save():
    dclab.PolygonFilter.clear_all_filters()

    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)

    # Add polygon filter
    pf = dclab.PolygonFilter(filename=tf)

    _fd, tf2 = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf2, "w") as fd:
        fd.write(filter_data)
        pf.save(tf2, ret_fobj=True)
        pf2 = dclab.PolygonFilter(filename=tf2)
        assert np.allclose(pf.points, pf2.points)

    _fd, tf3 = tempfile.mkstemp(prefix="dclab_polgyon_test")
    dclab.PolygonFilter.save_all(tf3)
    pf.save(tf3, ret_fobj=False)

    # ensure backwards compatibility: the names of the
    # three filters should be the same
    names = dclab.polygon_filter.get_polygon_filter_names()
    assert len(names) == 2
    assert names.count(names[0]) == 2

    try:
        os.remove(tf)
        os.remove(tf2)
        os.remove(tf3)
    except OSError:
        pass


@pytest.mark.filterwarnings('ignore::dclab.polygon_filter.'
                            + 'FilterIdExistsWarning')
def test_save_multiple():
    dclab.PolygonFilter.clear_all_filters()

    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)

    # Add polygon filter
    pf = dclab.PolygonFilter(filename=tf)

    _fd, tf2 = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf2, "a") as fd:
        pf.save(fd)
        pf2 = dclab.PolygonFilter(filename=tf2)
        assert np.allclose(pf.points, pf2.points)

    try:
        os.remove(tf)
        os.remove(tf2)
    except OSError:
        pass


def test_state():
    dclab.PolygonFilter.clear_all_filters()

    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)

    # Add polygon filter
    pf = dclab.PolygonFilter(filename=tf)

    state = pf.__getstate__()
    assert state["name"] == "polygon filter 0"
    assert state["axis x"] == "aspect"
    assert state["axis y"] == "tilt"
    assert np.allclose(state["points"][0][0], 6.344607717656481e-03)
    assert np.allclose(state["points"][3][1], 1.015706806282723e-03)
    assert not state["inverted"]

    state["name"] = "peter"
    state["axis x"] = "tilt"
    state["axis y"] = "aspect"
    state["points"][0][0] = 1
    state["inverted"] = True

    pf.__setstate__(state)

    assert pf.name == "peter"
    assert pf.axes[0] == "tilt"
    assert pf.axes[1] == "aspect"
    assert np.allclose(pf.points[0, 0], 1)
    assert pf.inverted

    try:
        os.remove(tf)
    except OSError:
        pass


@pytest.mark.filterwarnings('ignore::dclab.polygon_filter.'
                            + 'FilterIdExistsWarning')
def test_unique_id():
    dclab.PolygonFilter.clear_all_filters()
    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)

    # Add polygon filter
    pf = dclab.PolygonFilter(filename=tf, unique_id=2)
    pf2 = dclab.PolygonFilter(filename=tf, unique_id=2)
    assert pf.unique_id != pf2.unique_id
    dclab.PolygonFilter.clear_all_filters()

    try:
        os.remove(tf)
    except OSError:
        pass


@pytest.mark.filterwarnings('ignore::dclab.polygon_filter.'
                            + 'FilterIdExistsWarning')
def test_with_rtdc_data_set():
    dclab.PolygonFilter.clear_all_filters()
    ddict = example_data_dict(size=821, keys=["aspect", "tilt"])
    ds = dclab.new_dataset(ddict)

    # save polygon data
    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data)
    pf = dclab.PolygonFilter(filename=tf)
    pf2 = dclab.PolygonFilter(filename=tf)

    ds.polygon_filter_add(pf)
    ds.polygon_filter_add(1)

    ds.polygon_filter_rm(0)
    ds.polygon_filter_rm(pf2)

    dclab.PolygonFilter.clear_all_filters()
    try:
        os.remove(tf)
    except OSError:
        pass


def test_wrong_load_key():
    dclab.PolygonFilter.clear_all_filters()

    # save polygon data
    _fd, tf = tempfile.mkstemp(prefix="dclab_polgyon_test")
    with open(tf, "w") as fd:
        fd.write(filter_data + "peter=4\n")

    try:
        dclab.PolygonFilter(filename=tf)
    except KeyError:
        pass
    else:
        raise ValueError("_load should not accept unknown key!")
    dclab.PolygonFilter.clear_all_filters()
    try:
        os.remove(tf)
    except OSError:
        pass


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""
.. versionadded:: 0.38.0
"""
import hashlib

from ..feat_anc_core import AncillaryFeature
from ...util import obj2bytes

from . import modc


class MachineLearningFeature(AncillaryFeature):
    def __init__(self, feature_name, dc_model, modc_path=None):
        """A user-defined machine-learning feature

        Parameters
        ----------
        feature_name: str
            name of the ML feature score (starts with `ml_score_`)
        dc_model: dclab.rtdc_dataset.feat_anc_ml.ml_model.BaseModel
            ML model to register
        modc_path: str or Path
            path to the original .modc file (if applicable)

        Notes
        -----
        `MachineLearningFeature` inherits from :class:`AncillaryFeature
        <dclab.rtdc_dataset.feat_anc_core.ancillary_feature.AncillaryFeature>`.
        """
        if (not feature_name.startswith("ml_score_")
                or len(feature_name) != len("ml_score_123")):
            raise ValueError("Feature name for MachineLearning Feature must be"
                             + "in the form 'ml_score_xyz',"
                             + f"got '{feature_name}'!")
        # Make sure this MachineLearningFeature has not already been
        # registered (for normal features this is ok, but here we want
        # to avoid any possible ambiguity).
        for af in AncillaryFeature.features:
            if (isinstance(af, MachineLearningFeature)
                    and af.feature_name == feature_name):
                raise ValueError("Cannot register two MachineLearningFeatures "
                                 + f"for the same feature '{feature_name}'!")

        self.feature_name = feature_name
        self.modc_path = modc_path
        self.ml_feature_info = self._process_ml_info(dc_model.info)

        # Instantiate the super class
        super(MachineLearningFeature, self).__init__(
            feature_name=feature_name,
            method=dc_model.predict,
            req_features=dc_model.inputs,
            data=dc_model,
            identifier=self.ml_feature_info["identifier"],
        )

    def _process_ml_info(self, original_info):
        """Return dictionary with all relevant info for this instance
        """
        fidx = original_info["output features"].index(self.feature_name)

        # determine feature label
        if ("output labels" in original_info
                and original_info["output labels"][fidx]):
            label = original_info["output labels"][fidx]
        else:
            label = f"ML feature {self.feature_name}"

        md5hasher = hashlib.md5()
        md5hasher.update(original_info.get("sha256", "").encode())
        md5hasher.update(original_info.get("description", "").encode())
        md5hasher.update(original_info.get("long description", "").encode())
        md5hasher.update(original_info.get("date", "").encode())
        md5hasher.update(self.feature_name.encode("utf-8"))
        for feat in original_info["input features"]:
            md5hasher.update(feat.encode("utf-8"))
        identifier = md5hasher.hexdigest()

        feature_info = {
            "date": original_info.get("date", "no date"),
            "description": original_info.get(
                "description", "No description provided"),
            "long description": original_info.get(
                "long description", "No long description provided."),
            "feature name": self.feature_name,
            "feature label": label,
            "features required": original_info["input features"],
            "scalar feature": True,
            "modc path": self.modc_path,
            "identifier": identifier,
        }

        return feature_info

    def hash(self, rtdc_ds):
        """Used for identifying an ancillary ML computation

        The required features, the used configuration keys/values, and
        the return value of the requirement function are hashed. In
        addition, the ML model's info dictionary is hashed.
        """
        super_hash = super(MachineLearningFeature, self).hash(rtdc_ds)
        hasher = hashlib.md5()
        hasher.update(obj2bytes(super_hash))
        hasher.update(obj2bytes(self.data.info))
        return hasher.hexdigest()


def load_ml_feature(modc_path):
    """Find and load MachineLearningFeature(s) from a .modc file

    Parameters
    ----------
    modc_path: str or Path
        pathname to a .modc file

    Returns
    -------
    ml_list: list of MachineLearningFeature
        list of MachineLearningFeature instances loaded from `modc_path`

    See Also
    --------
    MachineLearningFeature: class handling the plugin feature information
    """
    dc_models = modc.load_modc(modc_path)

    mlf_list = []
    for dc_model in dc_models:
        for feat in dc_model.outputs:
            mlf_list.append(MachineLearningFeature(feat, dc_model, modc_path))

    return mlf_list


def remove_all_ml_features():
    """Convenience function for removing all `MachineLearningFeature` instances

    See Also
    --------
    remove_ml_feature: remove a single `MachineLearningFeature` instance
    """
    for ml_instance in reversed(MachineLearningFeature.features):
        if isinstance(ml_instance, MachineLearningFeature):
            remove_ml_feature(ml_instance)


def remove_ml_feature(ml_instance):
    """Convenience function for removing a `MachineLearningFeature` instance

    Parameters
    ----------
    ml_instance: MachineLearningFeature
        The `MachineLearningFeature` instance to be removed from dclab

    Raises
    ------
    TypeError
        If the `ml_instance` is not a `MachineLearningFeature` instance
    """
    if isinstance(ml_instance, MachineLearningFeature):
        MachineLearningFeature.feature_names.remove(ml_instance.feature_name)
        MachineLearningFeature.features.remove(ml_instance)
    else:
        raise TypeError(f"Type {type(ml_instance)} should be an instance "
                        f"of MachineLearningFeature; got '{ml_instance}'!")



"""Very basic docstrings for argparse.ArgumentParser

Usage
-----
To document a command, simply pass the location of a function that
returns an argparse.ArgumentParser instance.

   .. simple_argparse::
       :module: module.submodule
       :func: returns_argument_parser
       :prog: command-line-tool-name

Features:
- prints parameters as table instead of adding a new heading
- no additional modules needed (e.g. CommonMark)

Downsides:
- not as many functionalities as 'sphinxarg.ext'
- makes use of private function argparse.ArgumentParser._actions

Changelog:
0.3 (2020-02-14)
 - change "positional" to "required" arguments
 - check the "required" property to include keyword arguements
   that are required
0.2 (2019-07-12)
 - remove tables, use bullet points (improve readability for long
   docstrings)
 - correct usage of "metavar" instead of "dest"
0.1
 - good for short docstrings
"""
import importlib

from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive, directives
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes


class ArgparseDirective(Directive):
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True

    option_spec = {
        'module': directives.unchanged,
        'func': directives.unchanged,
        'prog': directives.unchanged,
    }

    def generate_rst(self):
        rst = []

        # get parser instance
        parser = self.get_parser()

        # description
        rst.append(parser.description)
        rst.append("")

        # usage
        usage = parser.format_usage()

        rst.append(".. code::")
        rst.append("")
        rst.append("    {}".format(usage))
        rst.append("")

        # parameters
        pos, opt = self.get_actions()

        if pos:
            rst.append("")
            rst.append("**required arguments:**")
            rst.append("")
            for item in pos:
                rst.append("- ``{}``".format(item[0]))
                rst.append("  {}".format(item[1]))
            rst.append("")

        if opt:
            rst.append("")
            rst.append("**optional arguments:**")
            rst.append("")
            for item in opt:
                if isinstance(item[3], bool):
                    if item[3]:
                        default = " *(enabled by default)*"
                    else:
                        default = " *(disabled by default)*"
                elif item[3] != "":
                    default = " *(default: {})*".format(item[3])
                else:
                    default = ""
                rst.append("- ``{}`` {} ".format(item[1], default))
                rst.append("  {}".format(item[2]))
            rst.append("")

        return rst

    def get_parser(self):
        """This will import the module and get the parser"""
        mod = importlib.import_module(self.options["module"])
        func = getattr(mod, self.options["func"])
        parser = func()
        parser.prog = self.options["prog"]
        return parser

    def get_actions(self):
        parser = self.get_parser()
        actions = parser._actions[1:]  # ignore help action
        positional = []
        optional = []
        for ac in actions:
            if ac.option_strings and not ac.required:
                optional.append([
                    ac.metavar,
                    ", ".join(ac.option_strings),
                    ac.help,
                    ac.default if ac.default is not None else "",
                ])
            else:
                positional.append([
                    ac.metavar,
                    ac.help,
                ])
        return positional, optional

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


def setup(app):
    app.add_directive('simple_argparse', ArgparseDirective)
    return {'version': '0.3'}   # identifies the version of our extension



# flake8: noqa: F401
from .ml_feature import (
    MachineLearningFeature, load_ml_feature, remove_all_ml_features)
from .modc import save_modc, load_modc
from . import ml_model

from . import hook_tensorflow



"""Repack (similar to h5repack) .rtdc files"""
import argparse
import pathlib

from ..rtdc_dataset import new_dataset, RTDCWriter
from .. import definitions as dfn

from . import common


def repack(path_in=None, path_out=None, strip_logs=False, check_suffix=True):
    """Repack/recreate an .rtdc file, optionally stripping the logs"""
    if path_in is None and path_out is None:
        parser = repack_parser()
        args = parser.parse_args()
        path_in = args.input
        path_out = args.output
        strip_logs = args.strip_logs

    allowed_input_suffixes = [".rtdc"]
    if not check_suffix:
        allowed_input_suffixes.append(pathlib.Path(path_in).suffix)

    path_in, path_out, path_temp = common.setup_task_paths(
        path_in, path_out, allowed_input_suffixes=allowed_input_suffixes)

    with new_dataset(path_in) as ds, RTDCWriter(path_temp, mode="reset") as hw:
        # write metadata first (to avoid resetting software version)
        # only export configuration meta data (no analysis-related config)
        meta = {}
        for sec in list(dfn.CFG_METADATA.keys()) + ["user"]:
            if sec in ds.config:
                meta[sec] = ds.config[sec].copy()

        hw.store_metadata(meta)

        if not strip_logs:
            for name in ds.logs:
                hw.store_log(name, ds.logs[name])

        # write features
        for feat in ds.features_innate:
            hw.store_feature(feat, ds[feat])

    # Finally, rename temp to out
    path_temp.rename(path_out)


def repack_parser():
    descr = "Repack an .rtdc file. The difference to dclab-compress " \
            + "is that no logs are added. Other logs can optionally be " \
            + "stripped away. Repacking also gets rid of old clutter " \
            + "data (e.g. previous metadata stored in the HDF5 file)."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.rtdc file)')
    parser.add_argument('output',  metavar="OUTPUT", type=str,
                        help='Output path (.rtdc file)')
    parser.add_argument('--strip-logs',
                        dest='strip_logs',
                        action='store_true',
                        help='Do not copy any logs to the output file.')
    parser.set_defaults(strip_logs=False)
    return parser



"""visualization of dclab definitions

Usage
-----
Directives:

List of features with arguments "all", "scalar", or "non-scalar"

   .. dclab_features::

List of configuration keys with arguments "analysis", "metadata", or a
configuration key section (e.g. "experiment").

   .. dclab_config::

"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

from dclab import definitions as dfn


class Base(Directive):
    required_arguments = 1
    optional_arguments = 0

    def generate_rst(self):
        pass

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


class Config(Base):
    def generate_rst(self):
        which = self.arguments[0]
        rst = []

        if which == "analysis":
            cfg = dfn.CFG_ANALYSIS
        elif which == "metadata":
            cfg = dfn.CFG_METADATA
        else:
            cfg = dfn.CFG_ANALYSIS.copy()
            cfg.update(dfn.CFG_METADATA)
            cfg = {which: cfg[which]}

        for key in sorted(cfg.keys()):
            rst.append("")
            rst.append(".. csv-table::")
            rst.append(
                "    :header: {}, parsed, description [units]".format(key))
            rst.append("    :widths: 30, 10, 60")
            rst.append("    :delim: tab")
            rst.append("")

            for item in sorted(cfg[key]):
                if item[1] is str:
                    ref = ":class:`str`"
                elif item[1] is float:
                    ref = ":class:`float`"
                else:
                    ref = ":func:`{f} <dclab.definitions" \
                          + f".meta_parse.{item[1].__name__}>`"
                rst.append("    {}\t {}\t {}".format(item[0],
                                                     ref,
                                                     item[2]))

            rst.append("")

        return rst


class Features(Base):
    def generate_rst(self):
        which = self.arguments[0]
        rst = []

        if which == "all":
            feats = sorted(dfn.FEATURES_SCALAR + dfn.FEATURES_NON_SCALAR)
        elif which == "scalar":
            feats = sorted(dfn.FEATURES_SCALAR)
        elif which == "non-scalar":
            feats = sorted(dfn.FEATURES_NON_SCALAR)

        rst.append(".. csv-table::")
        rst.append(
            "    :header: {} features, description [units]".format(which))
        rst.append("    :widths: 2, 7")
        rst.append("    :delim: tab")
        rst.append("")

        for item in feats:
            rst.append("    {}\t {}".format(item[0], item[1]))

        rst.append("")

        return rst


def setup(app):
    app.add_directive('dclab_features', Features)
    app.add_directive('dclab_config', Config)
    return {'version': '0.1'}   # identifies the version of our extension



#cython: cdivision=True
#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False
#cython: language_level=3
import numpy as np


cdef inline double _get_fraction(double from_value, double to_value,
                                 double level):
    if (to_value == from_value):
        return 0
    return ((level - from_value) / (to_value - from_value))


def iterate_and_store(double[:, :] array,
                      double level, Py_ssize_t vertex_connect_high):
    """Iterate across the given array in a marching-squares fashion,
    looking for segments that cross 'level'. If such a segment is
    found, its coordinates are added to a growing list of segments,
    which is returned by the function.  if vertex_connect_high is
    nonzero, high-values pixels are considered to be face+vertex
    connected into objects; otherwise low-valued pixels are.

    """
    if array.shape[0] < 2 or array.shape[1] < 2:
        raise ValueError("Input array must be at least 2x2.")

    cdef list arc_list = []
    cdef Py_ssize_t n

    # The plan is to iterate a 2x2 square across the input array. This means
    # that the upper-left corner of the square needs to iterate across a
    # sub-array that's one-less-large in each direction (so that the square
    # never steps out of bounds). The square is represented by four pointers:
    # ul, ur, ll, and lr (for 'upper left', etc.). We also maintain the current
    # 2D coordinates for the position of the upper-left pointer. Note that we
    # ensured that the array is of type 'double' and is C-contiguous (last
    # index varies the fastest).

    # Current coords start at 0,0.
    cdef Py_ssize_t[2] coords
    coords[0] = 0
    coords[1] = 0

    # Calculate the number of iterations we'll need
    cdef Py_ssize_t num_square_steps = (array.shape[0] - 1) \
                                        * (array.shape[1] - 1)

    cdef unsigned char square_case = 0
    cdef tuple top, bottom, left, right
    cdef double ul, ur, ll, lr
    cdef Py_ssize_t r0, r1, c0, c1

    for n in range(num_square_steps):
        # There are sixteen different possible square types, diagramed below.
        # A + indicates that the vertex is above the contour value, and a -
        # indicates that the vertex is below or equal to the contour value.
        # The vertices of each square are:
        # ul ur
        # ll lr
        # and can be treated as a binary value with the bits in that order. Thus
        # each square case can be numbered:
        #  0--   1+-   2-+   3++   4--   5+-   6-+   7++
        #   --    --    --    --    +-    +-    +-    +-
        #
        #  8--   9+-  10-+  11++  12--  13+-  14-+  15++
        #   -+    -+    -+    -+    ++    ++    ++    ++
        #
        # The position of the line segment that cuts through (or
        # doesn't, in case 0 and 15) each square is clear, except in
        # cases 6 and 9. In this case, where the segments are placed
        # is determined by vertex_connect_high.  If
        # vertex_connect_high is false, then lines like \\ are drawn
        # through square 6, and lines like // are drawn through square
        # 9.  Otherwise, the situation is reversed.
        # Finally, recall that we draw the lines so that (moving from tail to
        # head) the lower-valued pixels are on the left of the line. So, for
        # example, case 1 entails a line slanting from the middle of the top of
        # the square to the middle of the left side of the square.

        r0, c0 = coords[0], coords[1]
        r1, c1 = r0 + 1, c0 + 1

        ul = array[r0, c0]
        ur = array[r0, c1]
        ll = array[r1, c0]
        lr = array[r1, c1]

        # now in advance the coords indices
        if coords[1] < array.shape[1] - 2:
            coords[1] += 1
        else:
            coords[0] += 1
            coords[1] = 0


        square_case = 0
        if (ul > level): square_case += 1
        if (ur > level): square_case += 2
        if (ll > level): square_case += 4
        if (lr > level): square_case += 8

        if (square_case != 0 and square_case != 15):
            # only do anything if there's a line passing through the
            # square. Cases 0 and 15 are entirely below/above the contour.

            top = r0, c0 + _get_fraction(ul, ur, level)
            bottom = r1, c0 + _get_fraction(ll, lr, level)
            left = r0 + _get_fraction(ul, ll, level), c0
            right = r0 + _get_fraction(ur, lr, level), c1

            if (square_case == 1):
                # top to left
                arc_list.append(top)
                arc_list.append(left)
            elif (square_case == 2):
                # right to top
                arc_list.append(right)
                arc_list.append(top)
            elif (square_case == 3):
                # right to left
                arc_list.append(right)
                arc_list.append(left)
            elif (square_case == 4):
                # left to bottom
                arc_list.append(left)
                arc_list.append(bottom)
            elif (square_case == 5):
                # top to bottom
                arc_list.append(top)
                arc_list.append(bottom)
            elif (square_case == 6):
                if vertex_connect_high:
                    arc_list.append(left)
                    arc_list.append(top)

                    arc_list.append(right)
                    arc_list.append(bottom)
                else:
                    arc_list.append(right)
                    arc_list.append(top)
                    arc_list.append(left)
                    arc_list.append(bottom)
            elif (square_case == 7):
                # right to bottom
                arc_list.append(right)
                arc_list.append(bottom)
            elif (square_case == 8):
                # bottom to right
                arc_list.append(bottom)
                arc_list.append(right)
            elif (square_case == 9):
                if vertex_connect_high:
                    arc_list.append(top)
                    arc_list.append(right)

                    arc_list.append(bottom)
                    arc_list.append(left)
                else:
                    arc_list.append(top)
                    arc_list.append(left)

                    arc_list.append(bottom)
                    arc_list.append(right)
            elif (square_case == 10):
                # bottom to top
                arc_list.append(bottom)
                arc_list.append(top)
            elif (square_case == 11):
                # bottom to left
                arc_list.append(bottom)
                arc_list.append(left)
            elif (square_case == 12):
                # lef to right
                arc_list.append(left)
                arc_list.append(right)
            elif (square_case == 13):
                # top to right
                arc_list.append(top)
                arc_list.append(right)
            elif (square_case == 14):
                # left to top
                arc_list.append(left)
                arc_list.append(top)

    return arc_list




"""lme4: Linear mixed-effects models

We would like to quantify the difference between human skeletal stem cells
(SSC) and the human osteosarcoma cell line MG-63 (which is often used as a
model system for SSCs) using a likelihood ratio test based on LMM.

This example illustrates a basic LMM analysis. The data are loaded
from DCOR (:cite:`FigshareLMM`, `DCOR:figshare-11662773-v2
<https://dcor.mpl.mpg.de/dataset/figshare-11662773-v2>`_).
We treat SSC as our "treatment" and MG-63 as our "control" group.
These are just names that remind us that we are comparing one type of sample
against another type.

We are interested in the p-value, which is 0.01256 for
deformation. We repeat the analysis with area (0.0002183) and Young's
modulus (0.0002771). The p-values indicate that MG-63 (mean elastic
modulus 1.26 kPa) cells are softer than SSCs (mean elastic modulus 1.54 kPa).
The figure reproduces the last subplot of figure 6b im :cite:`Herbig2018`.
"""
import dclab
from dclab import lme4

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt


# https://dcor.mpl.mpg.de/dataset/figshare-11662773-v2
# SSC_16uls_rep1_20150611.rtdc
ds_ssc_rep1 = dclab.new_dataset("86cc5a47-364b-cf58-f9e3-cc114dd38e55")
# SSC_16uls_rep2_20150611.rtdc
ds_ssc_rep2 = dclab.new_dataset("ab95c914-0311-6a46-4eba-8fabca7d27d6")
# MG63_pure_16uls_rep1_20150421.rtdc
ds_mg63_rep1 = dclab.new_dataset("42cb33d4-2f7c-3c22-88e1-b9102d64d7e9")
# MG63_pure_16uls_rep2_20150422.rtdc
ds_mg63_rep2 = dclab.new_dataset("a4a98fcb-1de1-1048-0efc-b0a84d4ab32e")
# MG63_pure_16uls_rep3_20150422.rtdc
ds_mg63_rep3 = dclab.new_dataset("0a8096ce-ea7a-e36d-1df3-42c7885cd71c")

datasets = [ds_ssc_rep1, ds_ssc_rep2, ds_mg63_rep1, ds_mg63_rep2, ds_mg63_rep3]
for ds in datasets:
    # perform filtering
    ds.config["filtering"]["area_ratio min"] = 0
    ds.config["filtering"]["area_ratio max"] = 1.05
    ds.config["filtering"]["area_um min"] = 120
    ds.config["filtering"]["area_um max"] = 550
    ds.config["filtering"]["deform min"] = 0
    ds.config["filtering"]["deform max"] = 0.1
    ds.apply_filter()
    # enable computation of Young's modulus
    ds.config["calculation"]["emodulus lut"] = "LE-2D-FEM-19"
    ds.config["calculation"]["emodulus medium"] = "CellCarrier"
    ds.config["calculation"]["emodulus temperature"] = 23.0

# setup lme4 analysis
rlme4 = lme4.Rlme4(model="lmer")
rlme4.add_dataset(ds_ssc_rep1, group="treatment", repetition=1)
rlme4.add_dataset(ds_ssc_rep2, group="treatment", repetition=2)
rlme4.add_dataset(ds_mg63_rep1, group="control", repetition=1)
rlme4.add_dataset(ds_mg63_rep2, group="control", repetition=2)
rlme4.add_dataset(ds_mg63_rep3, group="control", repetition=3)

# perform analysis for deformation
for feat in ["area_um", "deform", "emodulus"]:
    res = rlme4.fit(feature=feat)
    print("Results for {}:".format(feat))
    print("  p-value", res["anova p-value"])
    print("  mean of MG-63", res["fixed effects intercept"])
    print("  fixed effect size", res["fixed effects treatment"])

# prepare for plotting
df = pd.DataFrame()
for ds in datasets:
    group = ds.config["experiment"]["sample"].split()[0]
    rep = ds.config["experiment"]["sample"].split()[-1]
    dfi = pd.DataFrame.from_dict(
        {"area_m": ds["area_um"][ds.filter.all],
         "deform": ds["deform"][ds.filter.all],
         "emodulus": ds["emodulus"][ds.filter.all],
         "group and repetition": [group + " " + rep] * ds.filter.all.sum(),
         "group": [group] * ds.filter.all.sum(),
         })
    df = df.append(dfi)

# plot
fig = plt.figure(figsize=(8, 5))
ax = sns.boxplot(x="group and repetition", y="emodulus", data=df, hue="group")
# note that `res` is still the result for "emodulus"
numstars = sum([res["anova p-value"] < .05,
                res["anova p-value"] < .01,
                res["anova p-value"] < .001,
                res["anova p-value"] < .0001])
# significance bars
h = .1
y1 = 6
y2 = 4.2
y3 = 6.2
ax.plot([-.5, -.5, 1, 1], [y1, y1+h, y1+h, y1], lw=1, c="k")
ax.plot([2, 2, 4.5, 4.5], [y2, y2+h, y2+h, y2], lw=1, c="k")
ax.plot([.25, .25, 3.25, 3.25], [y1+h, y1+2*h, y1+2*h, y2+h], lw=1, c="k")
ax.text(2, y3, "*"*numstars, ha='center', va='bottom', color="k")
ax.set_ylim(0, 7)

plt.tight_layout()
plt.show()



"""Utility methods"""
import functools
import hashlib
import pathlib
import warnings

import h5py
import numpy as np


def hashfile(fname, blocksize=65536, count=0, constructor=hashlib.md5,
             hasher_class=None):
    """Compute md5 hex-hash of a file

    Parameters
    ----------
    fname: str or pathlib.Path
        path to the file
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    hasher_class: callable
        deprecated, see use `constructor` instead
    constructor: callable
        hash algorithm constructor
    """
    if hasher_class is not None:
        warnings.warn("The `hasher_class` argument is deprecated, please use "
                      "`constructor` instead.")
        constructor = hasher_class
    path = pathlib.Path(fname).resolve()
    path_stat = path.stat()
    return _hashfile_cached(
        path=path,
        path_stats=(path_stat.st_mtime_ns, path_stat.st_size),
        blocksize=blocksize,
        count=count,
        constructor=constructor
    )


@functools.lru_cache(maxsize=100)
def _hashfile_cached(path, path_stats, blocksize=65536, count=0,
                     constructor=hashlib.md5):
    """Cached hashfile using stat tuple as cache

    This is a privat function. Please use `hashfile` instead!

    Parameters
    ----------
    path: pathlib.Path
        path to the file to be hashed
    path_stats: tuple
        tuple that contains information about the size and the
        modification time of `path`. This must be specified,
        so that caching of the result is done properly in case the user
        modified `path` (this function is wrapped with
        functools.lru_cache)
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    constructor: callable
        hash algorithm constructor
    """
    assert path_stats, "We need stat for validating the cache"
    hasher = constructor()
    with path.open('rb') as fd:
        buf = fd.read(blocksize)
        ii = 0
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
            ii += 1
            if count and ii == count:
                break
    return hasher.hexdigest()


def hashobj(obj):
    """Compute md5 hex-hash of a Python object"""
    return hashlib.md5(obj2bytes(obj)).hexdigest()


def obj2bytes(obj):
    """Bytes representation of an object for hashing"""
    if isinstance(obj, str):
        return obj.encode("utf-8")
    elif isinstance(obj, pathlib.Path):
        return obj2bytes(str(obj))
    elif isinstance(obj, (bool, int, float)):
        return str(obj).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tobytes()
    elif isinstance(obj, tuple):
        return obj2bytes(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2bytes(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2bytes(sorted(obj.items()))
    elif hasattr(obj, "identifier"):
        return obj2bytes(obj.identifier)
    elif isinstance(obj, h5py.Dataset):
        return obj2bytes(obj[0])
    else:
        raise ValueError("No rule to convert object '{}' to string.".
                         format(obj.__class__))



import sys

import h5py
import numpy as np
import pytest

import dclab.rtdc_dataset.config
from dclab.rtdc_dataset import check, check_dataset, fmt_tdms, new_dataset, \
    RTDCWriter

from helper_methods import example_data_dict, retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_basic():
    h5path = retrieve_data("fmt-hdf5_fl_2017.zip")
    viol, aler, info = check_dataset(h5path)
    # Features: Unknown key 'ncells'
    # Metadata: Mismatch [imaging] 'roi size x' and feature image (50 vs 90)
    # Metadata: Mismatch [imaging] 'roi size y' and feature image (90 vs 50)
    # Metadata: Missing key [fluorescence] 'channels installed'
    # Metadata: Missing key [fluorescence] 'laser count'
    # Metadata: Missing key [fluorescence] 'lasers installed'
    # Metadata: Missing key [fluorescence] 'samples per event'
    # Metadata: fluorescence channel count inconsistent
    assert len(viol) == 8
    # "HDF5: '/image': attribute 'CLASS' should be fixed-length ASCII string",
    # "HDF5: '/image': attribute 'IMAGE_SUBCLASS' should be fixed-length ...
    # "HDF5: '/image': attribute 'IMAGE_VERSION' should be fixed-length ...
    # "Metadata: Flow rates don't add up (sh 0.6 + sam 0.1 != channel 0.16)",
    # "Metadata: Flow rates don't add up (sh 0.6 + sam 0.1 != channel 0.16)",
    # "Metadata: Flow rates don't add up (sh 0.6 + sam 0.1 != channel 0.16)",
    # "Metadata: Missing key [fluorescence] 'channel 1 name'",
    # "Metadata: Missing key [fluorescence] 'channel 2 name'",
    # "Metadata: Missing key [fluorescence] 'channel 3 name'",
    # "Metadata: Missing key [setup] 'identifier'",
    # "Metadata: Missing section 'online_contour'"
    # "UnknownConfigurationKeyWarning: Unknown key 'exposure time' ...",
    # "UnknownConfigurationKeyWarning: Unknown key 'flash current' ...",
    # "UserWarning: Type of confguration key [fluorescence]: sample rate ...",
    # "UserWarning: Type of confguration key [imaging]: roi position x ...",
    # "UserWarning: Type of confguration key [imaging]: roi position y ..."]
    assert len(aler) == 16
    assert "Data file format: hdf5" in info
    assert "Fluorescence: True" in info
    assert "Compression: Partial (1 of 25)" in info


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_complete():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    viol, aler, info = check_dataset(h5path)
    assert len(viol) == 0
    # [fluorescence]: sample rate should be <class 'numbers.Integral'>
    # [imaging]: roi position x should be <class 'numbers.Integral'>
    # [imaging]: roi position y should be <class 'numbers.Integral'>
    assert len(aler) == 3
    assert "Data file format: hdf5" in info
    assert "Fluorescence: True" in info


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_complete_user_metadata():
    """Setting any user metadata is allowed"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    with new_dataset(h5path) as ds:
        ds.config.update({"user": metadata})
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate)
        viol, aler, info = check_dataset(ds)
        assert len(viol) == 0
        assert len(aler) == 0
        assert "Data file format: hdf5" in info
        assert "Fluorescence: True" in info


def test_exact():
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_2fl-no-image_2017.zip")
    viol, aler, info = check_dataset(h5path)
    known_viol = [
        "Metadata: Missing key [fluorescence] 'channel count'",
        "Metadata: Missing key [fluorescence] 'channels installed'",
        "Metadata: Missing key [fluorescence] 'laser count'",
        "Metadata: Missing key [fluorescence] 'lasers installed'",
        "Metadata: Missing key [fluorescence] 'samples per event'",
    ]
    known_aler = [
        "Metadata: Missing key [fluorescence] 'channel 1 name'",
        "Metadata: Missing key [fluorescence] 'channel 2 name'",
        "Metadata: Missing key [online_contour] 'no absdiff'",
        "Metadata: Missing key [setup] 'identifier'",
        "Metadata: Missing key [setup] 'module composition'",
        "Negative value for feature(s): fl2_max",
    ]
    known_info = [
        'Compression: None',
        'Data file format: tdms',
        'Fluorescence: True',
    ]
    assert set(viol) == set(known_viol)
    assert set(aler) == set(known_aler)
    assert set(info) == set(known_info)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_icue():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check()
    assert cues[0] != cues[1]
    levels = check.ICue.get_level_summary(cues)
    assert levels["info"] >= 2
    # [fluorescence]: sample rate should be <class 'numbers.Integral'>
    # [imaging]: roi position x should be <class 'numbers.Integral'>
    # [imaging]: roi position y should be <class 'numbers.Integral'>
    assert levels["alert"] == 3
    assert levels["violation"] == 0
    assert cues[0].msg in cues[0].__repr__()


def test_ic_expand_section():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds1 = new_dataset(ddict)
    ds2 = new_dataset(ddict)
    with check.IntegrityChecker(ds1) as ic:
        cues1 = ic.check_metadata_missing(expand_section=True)
    with check.IntegrityChecker(ds2) as ic:
        cues2 = ic.check_metadata_missing(expand_section=False)
    assert len(cues1) > len(cues2)


def test_ic_feature_size_scalar():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ddict["bright_sd"] = np.linspace(10, 20, 1000)
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_feature_size()
    for cue in cues:
        if cue.category == "feature size":
            assert cue.msg == "Features: wrong event count: 'bright_sd' " \
                              + "(1000 of 8472)"
            break
    else:
        assert False


def test_ic_feature_size_trace():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ddict["trace"] = {"fl1_raw": [range(10)] * 1000}
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_feature_size()
    for cue in cues:
        if cue.category == "feature size":
            assert cue.msg == "Features: wrong event count: 'trace/fl1_raw'" \
                              + " (1000 of 8472)"
            break
    else:
        assert False


def test_ic_fl_metadata_channel_names():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform", "fl1_max"])
    ddict["trace"] = {"fl1_raw": [range(10)] * 1000}
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_metadata_channel_names()
    assert cues[0].category == "metadata missing"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "channel 1 name"


def test_ic_fl_metadata_channel_names_2():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["fluorescence"]["channel 1 name"] = "peter"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_metadata_channel_names()
    assert cues[0].category == "metadata invalid"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "channel 1 name"


def test_ic_fl_num_channels():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["fluorescence"]["channel count"] = 3
    ds.config["fluorescence"]["channel 1 name"] = "hans"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_num_channels()
    assert cues[0].category == "metadata wrong"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "channel count"


def test_ic_fl_num_lasers():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["fluorescence"]["laser count"] = 3
    ds.config["fluorescence"]["laser 1 lambda"] = 550
    ds.config["fluorescence"]["laser 1 power"] = 20
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_num_lasers()
    assert cues[0].category == "metadata wrong"
    assert cues[0].cfg_section == "fluorescence"
    assert cues[0].cfg_key == "laser count"


def test_ic_flow_rate_not_zero():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_metadata_bad_greater_zero()
    assert cues[0].category == "metadata wrong"
    assert cues[0].cfg_section == "setup"
    assert cues[0].cfg_key == "flow rate"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ic_fmt_hdf5_image1():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with h5py.File(h5path, "a") as h5:
        h5["events/image"].attrs["CLASS"] = "bad"
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image': attribute 'CLASS' should be " \
                          + "fixed-length ASCII string"
    assert cues[0].level == "alert"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(sys.version_info < (3, 0),
                    reason="requires python3 or higher")
def test_ic_fmt_hdf5_image2():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with h5py.File(h5path, "a") as h5:
        h5["events/image"].attrs["CLASS"] = np.string_("bad")
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image': attribute 'CLASS' should have " \
                          + "value 'b'IMAGE''"
    assert cues[0].level == "alert"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ic_fmt_hdf5_image3():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with h5py.File(h5path, "a") as h5:
        del h5["events/image"].attrs["CLASS"]
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image': missing attribute 'CLASS'"
    assert cues[0].level == "alert"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ic_fmt_hdf5_image_bg():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # add a fake image_bg column
    with h5py.File(h5path, "a") as h5:
        image_bg = h5["events"]["image"][:] // 2
        hw = RTDCWriter(h5)
        hw.store_feature("image_bg", image_bg)
        del h5["events/image_bg"].attrs["CLASS"]
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == "HDF5: '/image_bg': missing attribute 'CLASS'"
    assert cues[0].level == "alert"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ic_fmt_hdf5_logs():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    hw = RTDCWriter(h5path)
    hw.store_log("test", ["asdasd"*100])
    hw.store_log("M1_para.ini", ["asdasd"*100])
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_fmt_hdf5()
    assert len(cues) == 1
    assert cues[0].category == "format HDF5"
    assert cues[0].msg == 'Logs: test line 0 exceeds maximum line length 100'
    assert cues[0].level == "alert"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ic_metadata_bad():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    # Since version 0.35, metadata are checked in `Configuration` class
    with pytest.warns(dclab.rtdc_dataset.config.WrongConfigurationTypeWarning,
                      match="run index"):
        ds.config["experiment"]["run index"] = "1"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_metadata_bad()
    assert len(cues) == 0


def test_ic_metadata_choices_medium():
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["setup"]["medium"] = "honey"
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_metadata_choices()
    # changed in 0.29.1: medium can now be an arbitrary string
    # except for an empty string.
    assert len(cues) == 0


def test_check_metadata_hdf5_type_issue_139():
    """Check that chip region is lower-case"""
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:chip region"] = "Channel"
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.check_metadata_hdf5_type()
    assert len(cues) == 1
    assert cues[0].msg.count("channel")
    assert cues[0].msg.count("Channel")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.EmptyConfigurationKeyWarning')
def test_ic_metadata_empty_string():
    """Empty metadata values are ignored with a warning in dclab>0.33.2"""
    path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # add empty attribute
    with h5py.File(path, "r+") as h5:
        h5.attrs["setup:medium"] = ""
    ds = new_dataset(path)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_metadata_missing()
    assert cues[0].category == "metadata missing"
    assert cues[0].level == "violation"
    assert cues[0].cfg_section == "setup"
    assert cues[0].cfg_key == "medium"


def test_ic_fl_max():
    # Testing dataset with negative fl_max values
    ddict = example_data_dict(size=8472, keys=["fl1_max"])
    ddict["fl1_max"] -= min(ddict["fl1_max"]) + 1
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_max_positive()
    assert cues[0].level == "alert"
    assert cues[0].category == "feature data"

    # Testing dataset with fl_max values of 0.1
    ddict = example_data_dict(size=8472, keys=["fl1_max"])
    ddict["fl1_max"] -= min(ddict["fl1_max"]) - 0.1
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_max_positive()
    assert cues[0].level == "alert"
    assert cues[0].category == "feature data"

    # Testing dataset with fl_max values > 0.1
    ddict = example_data_dict(size=8472, keys=["fl1_max"])
    ddict["fl1_max"] -= min(ddict["fl1_max"]) - 1
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_max_positive()
    assert not cues


def test_ic_fl_max_ctc():
    # Testing dataset with negative fl_max_ctc values
    ddict = example_data_dict(size=8472, keys=["fl1_max_ctc"])
    ddict["fl1_max_ctc"] -= min(ddict["fl1_max_ctc"]) + 1
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_max_ctc_positive()
    assert cues[0].level == "alert"
    assert cues[0].category == "feature data"

    # Testing dataset with fl_max_ctc values of 0.1
    ddict = example_data_dict(size=8472, keys=["fl1_max_ctc"])
    ddict["fl1_max_ctc"] -= min(ddict["fl1_max_ctc"]) - 0.1
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_max_ctc_positive()
    assert cues[0].level == "alert"
    assert cues[0].category == "feature data"

    # Testing dataset with fl_max_ctc values > 0.1
    ddict = example_data_dict(size=8472, keys=["fl1_max_ctc"])
    ddict["fl1_max_ctc"] -= min(ddict["fl1_max_ctc"]) - 1
    ds = new_dataset(ddict)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_fl_max_ctc_positive()
    assert not cues


def test_ic_invalid_dataset():
    # Testing if IC throws NotImplementedError for hierarchy datasets
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds_child = new_dataset(ds)
    with check.IntegrityChecker(ds_child) as ic:
        with pytest.raises(NotImplementedError):
            ic.check()

    # Testing if IC throws NotImplementedError for raw-datasets with
    # applied filters
    ddict = example_data_dict(size=8472, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    ds.config["filtering"]["area_um max"] = 100
    ds.config["filtering"]["area_um min"] = 1
    ds.apply_filter()
    with check.IntegrityChecker(ds) as ic:
        with pytest.raises(NotImplementedError):
            ic.check()


def test_ic_sanity():
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    with h5py.File(h5path, "a") as h5:
        del h5["events"]["deform"]
        h5["events"]["deform"] = np.ones(100) * .1
    with check.IntegrityChecker(h5path) as ic:
        cues = ic.sanity_check()
    assert len(cues) == 1
    assert cues[0].category == "feature size"
    assert cues[0].msg.count("Sanity check failed:")
    assert cues[0].msg.count("deform")
    assert cues[0].level == "violation"


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="requires python3.6 or higher")
def test_invalid_medium():
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_minimal_2016.zip")
    para = h5path.with_name("M1_para.ini")
    cfg = para.read_text().split("\n")
    cfg.insert(3, "Buffer Medium = unknown_bad!")
    para.write_text("\n".join(cfg))
    viol, _, _ = check_dataset(h5path)
    # changed in 0.29.1: medium can now be an arbitrary string
    # except for an empty string.
    assert "Metadata: Invalid value [setup] medium: 'unknown_bad!'" not in viol


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_load_with():
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_minimal_2016.zip")
    known_aler = [
        "Metadata: Missing key [setup] 'flow rate sample'",
        "Metadata: Missing key [setup] 'flow rate sheath'",
        "Metadata: Missing key [setup] 'identifier'",
        "Metadata: Missing key [setup] 'module composition'",
        "Metadata: Missing key [setup] 'software version'",
    ]
    known_viol = [
        "Features: wrong event count: 'contour' (14 of 156)",
        "Features: wrong event count: 'mask' (14 of 156)",
        "Metadata: Missing key [setup] 'medium'",
    ]
    with new_dataset(h5path) as ds:
        viol, aler, _ = check_dataset(ds)
        assert set(viol) == set(known_viol)
        assert set(aler) == set(known_aler)


def test_missing_file():
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_2fl-no-image_2017.zip")
    h5path.with_name("M1_para.ini").unlink()
    try:
        check_dataset(h5path)
    except fmt_tdms.IncompleteTDMSFileFormatError:
        pass
    else:
        assert False


def test_ml_class():
    """Test score data outside boundary"""
    data = {"ml_score_001": [.1, 10, -10, 0.01, .89],
            "ml_score_002": [.2, .1, .4, 0, .4],
            }
    ds = new_dataset(data)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_ml_class()
        assert len(cues) == 1
        assert "ml_score_001" in cues[0].msg


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_no_fluorescence():
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_minimal_2016.zip")
    _, _, info = check_dataset(h5path)
    known_info = [
        'Compression: None',
        'Data file format: tdms',
        'Fluorescence: False',
    ]
    assert set(info) == set(known_info)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_online_polygon_filters():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # add an artificial online polygon filter
    with h5py.File(path, "a") as h5:
        # set soft filter to True
        h5.attrs["online_filter:area_um,deform soft limit"] = True
        # set filter values
        pf_name = "online_filter:area_um,deform polygon points"
        area_um = h5["events"]["area_um"]
        deform = h5["events"]["deform"]
        pf_points = np.array([
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform)],
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform) + np.std(deform)],
            [np.mean(area_um),
             np.mean(deform) + np.std(deform)],
        ])
        h5.attrs[pf_name] = pf_points

    # see if we can open the file without any error
    with check.IntegrityChecker(path) as ic:
        cues = [cc for cc in ic.check() if cc.level != "info"]
        # [imaging]: roi position x should be <class 'numbers.Integral'>
        # [imaging]: roi position y should be <class 'numbers.Integral'>
        assert len(cues) == 2


def test_online_polygon_filters_real_data():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")

    # see if we can open the file without any error
    with check.IntegrityChecker(path) as ic:
        cues = [cc for cc in ic.check() if cc.level != "info"]
        assert len(cues) == 0


@pytest.mark.parametrize("shape", [[3, 3], [2, 2], [1, 2], [10, 3]])
def test_online_polygon_filters_wrong_shape(shape):
    """Shape-In 2.3 supports online polygon filters (test for shape)"""
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(path, "a") as h5:
        for key in ["imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # add an artificial online polygon filter
    with h5py.File(path, "a") as h5:
        # set soft filter to True
        h5.attrs["online_filter:area_um,deform soft limit"] = True
        # set filter values
        pf_name = "online_filter:area_um,deform polygon points"
        pf_points = np.arange(shape[0]*shape[1]).reshape(*shape)
        h5.attrs[pf_name] = pf_points

    # see if we can open the file without any error
    with check.IntegrityChecker(path) as ic:
        cues = [cc for cc in ic.check() if cc.level != "info"]
        assert len(cues) == 1
        assert cues[0].category == "metadata wrong"
        assert cues[0].level == "violation"
        assert cues[0].cfg_section == "online_filter"
        assert cues[0].cfg_key == "area_um,deform polygon points"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("si_version", ["2.2.1.0", "2.2.1.0dev", "2.2.2.0dev",
                                        "2.2.2.0", "2.2.2.1", "2.2.2.2",
                                        "2.2.2.4", "2.2.2.4", "2.3.0.0"])
def test_shapein_issue3_bad_medium(si_version):
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = si_version
        h5.attrs["setup:medium"] = "CellCarrierB"
    ds = new_dataset(h5path)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_shapein_issue3_bad_medium()
        assert len(cues) == 1
        assert cues[0].cfg_key == "medium"
        assert cues[0].cfg_section == "setup"
        assert cues[0].category == "metadata wrong"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("si_version", ["2.2.0.3", "2.3.1.0"])
def test_shapein_issue3_bad_medium_control(si_version):
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = si_version
        h5.attrs["setup:medium"] = "CellCarrierB"
    ds = new_dataset(h5path)
    with check.IntegrityChecker(ds) as ic:
        cues = ic.check_shapein_issue3_bad_medium()
        assert len(cues) == 0


def test_temperature():
    # there are probably a million things wrong with this dataset, but
    # we are only looking for the temperature thing
    ddict = example_data_dict(size=8472, keys=["area_um", "deform", "temp"])
    ds = new_dataset(ddict)
    sstr = "Metadata: Missing key [setup] 'temperature', " \
           + "because the 'temp' feature is given"
    _, aler, _ = check_dataset(ds)
    assert sstr in aler


def test_wrong_samples_per_event():
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_2fl-no-image_2017.zip")
    with h5path.with_name("M1_para.ini").open("a") as fd:
        fd.write("Samples Per Event = 10\n")
    msg = "Metadata: wrong number of samples per event: fl1_median " \
          + "(expected 10, got 566)"
    viol, _, _ = check_dataset(h5path)
    assert msg in viol


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""skimage.measure.find_contours

This submodule is copied from skimage.measure to avoid the additional
dependency scikit-image in dclab.
"""
from ._find_contours import find_contours  # noqa: F401
from .pnpoly import points_in_poly  # noqa: F401



""" Test functions for loading contours
"""
import os
import tempfile
import pathlib
import warnings
import pytest

import numpy as np
import h5py

import dclab
from dclab.rtdc_dataset import new_dataset
import dclab.rtdc_dataset.config as dccfg
from test_rtdc_fmt_dcor import DCOR_AVAILABLE

from helper_methods import retrieve_data, example_data_dict


data_path = pathlib.Path(__file__).parent / "data"


def equals(a, b):
    """Compare objects with allclose"""
    if isinstance(a, (dict, dccfg.Configuration, dccfg.ConfigurationDict)):
        for key in a:
            assert key in b, "key not in b"
            assert equals(a[key], b[key])
    elif isinstance(a, (float, int)):
        if np.isnan(a):
            assert np.isnan(b)
        else:
            assert np.allclose(a, b), "a={} vs b={}".format(a, b)
    else:
        assert a == b, "a={} vs b={}".format(a, b)
    return True


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_config_basic():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert ds.config["imaging"]["roi size y"] == 96.


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_config_invalid_key():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        ds.config["setup"]["invalid_key"] = "picard"
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, dccfg.UnknownConfigurationKeyWarning)
        assert "invalid_key" in str(w[-1].message)


def test_config_save_load():
    pytest.importorskip("nptdms")
    # Download and extract data
    tdms_path = retrieve_data("fmt-tdms_minimal_2016.zip")
    ds = new_dataset(tdms_path)
    cfg_file = tempfile.mktemp(prefix="test_dclab_rtdc_config_")
    ds.config.save(cfg_file)
    loaded = dccfg.Configuration(files=[cfg_file])
    assert equals(loaded, ds.config)
    try:
        os.remove(cfg_file)
    except OSError:
        pass


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning")
@pytest.mark.parametrize("path", list(data_path.glob("fmt-hdf5_*")))
def test_config_to_json(path):
    ds = new_dataset(retrieve_data(path.name))
    ds.config.tojson()


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_config_update():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert ds.config["imaging"]["roi size y"] == 96.
    ds.config["calculation"].update({
        "crosstalk fl12": 0.1,
        "crosstalk fl13": 0.2,
        "crosstalk fl21": 0.3,
        "crosstalk fl23": 0.4,
        "crosstalk fl31": 0.5,
        "crosstalk fl32": 0.6,
    })

    assert ds.config["calculation"]["crosstalk fl12"] == 0.1
    assert ds.config["calculation"]["crosstalk fl13"] == 0.2
    assert ds.config["calculation"]["crosstalk fl21"] == 0.3
    assert ds.config["calculation"]["crosstalk fl23"] == 0.4
    assert ds.config["calculation"]["crosstalk fl31"] == 0.5
    assert ds.config["calculation"]["crosstalk fl32"] == 0.6


@pytest.mark.parametrize("section,key,descr",
                         [["online_filter",
                           "area_um,deform polygon points",
                           "Polygon (Area, Deformation)"],
                          ["online_filter",
                           "area_um,deform soft limit",
                           "Soft limit, polygon (Area, Deformation)"],
                          # sanity check
                          ["experiment",
                           "run index",
                           "Index of measurement run"],
                          ])
def test_get_config_value_descr(section, key, descr):
    """Descriptions of keys"""
    assert descr == dclab.dfn.get_config_value_descr(section, key)


def test_lower_case_conversion_issue_139():
    """Check that chip region is lower-case"""
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:chip region"] = "Channel"
    with new_dataset(h5path) as ds:
        assert ds.config["setup"]["chip region"] == "channel"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_allowed_key_types():
    """Check that the user config section keys only accept strings"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    # strings are allowed
    ds.config["user"]["a string"] = "a string"
    # all other types will raise a BadUserConfigurationKeyWarning
    with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
        ds.config["user"][True] = True
    with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
        ds.config["user"][23.5] = 23.5
    with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
        ds.config["user"][12] = 12
    with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
        ds.config["user"][(5, 12.3, False, "a word")] = "a word"
    with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
        ds.config["user"][[5, 12.3, False, "a word"]] = "a word"
    with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
        ds.config["user"][{"name": 12.3, False: "a word"}] = "a word"
    with pytest.warns(dccfg.BadUserConfigurationValueWarning):
        ds.config["user"]["name"] = None

    assert len(ds.config["user"]) == 1
    assert isinstance(ds.config["user"]["a string"], str)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_basic():
    """Add information to the user section of config"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    ds.config.update({"user": metadata})
    assert ds.config["user"] == metadata


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_clear1():
    """Clear information from the user section with `clear()` method"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    ds.config.update({"user": metadata})
    assert ds.config["user"] == metadata
    ds.config["user"].clear()
    assert ds.config["user"] == {}


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_clear2():
    """Clear information from the user section with empty dict"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    ds.config.update({"user": metadata})
    assert ds.config["user"] == metadata
    ds.config["user"] = {}
    assert ds.config["user"] == {}


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_different_value_types():
    """Check that the user config section values take different data types"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert ds.config["user"] == {}
    ds.config["user"]["channel A"] = True
    ds.config["user"]["channel B"] = 23.5
    ds.config["user"]["channel C"] = 12
    ds.config["user"]["channel D"] = "a string"
    ds.config["user"]["channel E"] = {"a key": 23}
    ds.config["user"]["channel F"] = [8, 55.4, True, "a word"]
    ds.config["user"]["channel G"] = (5, 12.3, False, "a word")

    assert len(ds.config["user"]) == 7
    assert isinstance(ds.config["user"]["channel A"], bool)
    assert isinstance(ds.config["user"]["channel B"], float)
    assert isinstance(ds.config["user"]["channel C"], int)
    assert isinstance(ds.config["user"]["channel D"], str)
    assert isinstance(ds.config["user"]["channel E"], dict)
    assert isinstance(ds.config["user"]["channel F"], list)
    assert isinstance(ds.config["user"]["channel G"], tuple)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_exists():
    """Check that the user config section exists"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert ds.config["user"] == {}
    # control: nonsense sections don't exist
    with pytest.raises(KeyError):
        ds.config["Oh I seem to have lost my key"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_and_overwrite():
    """Add information to the user section of config via dict.__setitem__"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    ds.config["user"]["some metadata"] = 42
    assert ds.config["user"] == {"some metadata": 42}
    metadata = {"more metadata": True}
    ds.config.update({"user": metadata})
    assert ds.config["user"] == {"some metadata": 42,
                                 "more metadata": True}
    # overwrite the previous keys and values
    ds.config["user"] = {}
    assert ds.config["user"] == {}


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_save_reload_empty_dict():
    """The 'user' config section as an empty dict will not save"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with new_dataset(h5path) as ds:
        ds.config.update({"user": {}})
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate)
    # nothing "user"-like is written to the HDF5 attributes
    with h5py.File(expath, "r") as h5:
        for ak in h5.attrs:
            assert not ak.startswith("user")
    # works for dclab because "user" added when checked
    with new_dataset(expath) as ds2:
        assert ds2.config["user"] == {}


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("user_config",
                         [
                             {"": "a"},
                             {"": "b"},
                             {" ": "c"},
                             {"   ": "peter"},
                             {"\t": "pan"},
                             {"\n": "hook"},
                             {"\r": "croc"},
                         ])
def test_user_section_set_save_reload_empty_key(user_config):
    """Empty or only-whitespace 'user' section keys not allowed"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with new_dataset(h5path) as ds:
        with pytest.warns(dccfg.BadUserConfigurationKeyWarning):
            ds.config.update({"user": user_config})


def test_user_section_set_save_reload_fmt_dict():
    """Check that 'user' section metadata works for RTDC_Dict"""
    # create temp directory for storing outputted file
    tpath = pathlib.Path(tempfile.mkdtemp())
    ddict = example_data_dict(size=67, keys=["area_um", "deform"])
    ds = new_dataset(ddict)
    metadata = {"some metadata": 42}
    ds.config.update({"user": metadata})
    assert ds.config["user"] == metadata
    # must add some metadata to the "experiment" key for loading with dclab
    ds.config["experiment"]["sample"] = "test"
    ds.config["experiment"]["run index"] = 1
    expath = tpath / "exported.rtdc"
    with expath as exp:
        ds.export.hdf5(exp, features=["deform", "area_um"])
    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert h5.attrs["user:some metadata"] == 42
    # check again with dclab
    with new_dataset(expath) as ds2:
        assert ds2.config["user"] == metadata


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not reachable!")
def test_user_section_set_save_reload_fmt_dcor():
    """Check that 'user' section metadata works for RTDC_DCOR"""
    # create temp directory for storing outputted file
    tpath = pathlib.Path(tempfile.mkdtemp())
    with new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        metadata = {"some metadata": 12}
        ds.config.update({"user": metadata})
        assert ds.config["user"] == metadata
        expath = tpath / "exported.rtdc"
        ds.export.hdf5(expath, features=["deform", "area_um"])
    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert h5.attrs["user:some metadata"] == 12
    # check it again with dclab
    with new_dataset(expath) as ds2:
        assert ds2.config["user"] == metadata


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_save_reload_fmt_hdf5_basic():
    """Check that 'user' section metadata works for RTDC_HDF5"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    with new_dataset(h5path) as ds:
        ds.config.update({"user": metadata})
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate)
    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert h5.attrs["user:channel area"] == 100.5
        assert h5.attrs["user:inlet"]
        assert h5.attrs["user:n_constrictions"] == 3
        assert h5.attrs["user:channel information"] == "other information"
    # now check again with dclab
    with new_dataset(expath) as ds2:
        assert ds2.config["user"] == metadata


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_save_reload_fmt_hdf5_containers():
    """Check that 'user' section metadata works for container data types"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    channel_area = [0, 100]
    inlet = (1, 20, 40)
    outlet = np.array(inlet)
    metadata = {"channel area": channel_area,
                "inlet": inlet,
                "outlet": outlet}
    with new_dataset(h5path) as ds:
        ds.config.update({"user": metadata})
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate)
    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert all(h5.attrs["user:channel area"] == channel_area)
        assert all(h5.attrs["user:inlet"] == inlet)
        assert all(h5.attrs["user:outlet"] == outlet)
    # now check again with dclab
    with new_dataset(expath) as ds2:
        for k1 in metadata:
            assert all(metadata[k1] == ds2.config["user"][k1])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_save_reload_fmt_hierarchy():
    """Check that 'user' section metadata works for RTDC_Hierarchy"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    with new_dataset(h5path) as ds:
        ds.config.update({"user": metadata})
        ch = new_dataset(ds)
        expath = h5path.with_name("exported.rtdc")
        ch.export.hdf5(expath, features=ch.features_innate)
    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert h5.attrs["user:channel area"] == 100.5
        assert h5.attrs["user:inlet"]
        assert h5.attrs["user:n_constrictions"] == 3
        assert h5.attrs["user:channel information"] == "other information"
    # now check again with dclab
    with new_dataset(expath) as ds2:
        assert ds2.config["user"] == metadata


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.export.'
                            + 'LimitingExportSizeWarning')
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.'
                            + 'InitialFrameMissingWarning')
def test_user_section_set_save_reload_fmt_tdms():
    """Check that 'user' section metadata works for RTDC_TDMS"""
    pytest.importorskip("nptdms")
    h5path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    metadata = {"channel area": 100.5,
                "inlet": True,
                "n_constrictions": 3,
                "channel information": "other information"}
    with new_dataset(h5path) as ds:
        ds.config.update({"user": metadata})
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate)
    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert h5.attrs["user:channel area"] == 100.5
        assert h5.attrs["user:inlet"]
        assert h5.attrs["user:n_constrictions"] == 3
        assert h5.attrs["user:channel information"] == "other information"
    # now check again with dclab
    with new_dataset(expath) as ds2:
        assert ds2.config["user"] == metadata


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_with_setitem():
    """Add information to the user section of config via dict.__setitem__"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    ds.config["user"]["some metadata"] = 42
    assert ds.config["user"] == {"some metadata": 42}


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_user_section_set_with_update():
    """Add information to the user section of config with .update"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    metadata = {"some metadata": 42}
    ds.config.update({"user": metadata})
    assert ds.config["user"] == {"some metadata": 42}

    metadata2 = {"channel information": "information"}
    ds.config["user"].update(metadata2)
    assert ds.config["user"] == {"some metadata": 42,
                                 "channel information": "information"}



import pathlib
import time

import h5py
import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.features.contour import get_contour, get_contour_lazily
from dclab.features.volume import get_volume

from scipy.ndimage import binary_fill_holes

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_contour_basic():
    ds1 = dclab.new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    # export all data except for contour data
    features = ds1.features
    features.remove("contour")
    dspath = pathlib.Path(ds1.path)
    tempout = dspath.parent / (dspath.name + "without_contour.rtdc")
    ds1.export.hdf5(tempout, features=features)
    ds2 = dclab.new_dataset(tempout)

    for ii in range(len(ds1)):
        cin = ds1["contour"][ii]
        cout = ds2["contour"][ii]
        # simple presence test
        for ci in cin:
            assert ci in cout
        # order
        for jj in range(1, len(cin)):
            c2 = np.roll(cin, jj, axis=0)
            if np.all(c2 == cout):
                break
        else:
            assert False, "contours not matching, check orientation?"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_artefact():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl-no-contour_2019.zip"))
    # This would raise a "dclab.features.contour.NoValidContourFoundError:
    # Event 1, No contour found!" in dclab version <= 0.22.1
    cont = ds["contour"][1]
    assert len(cont) == 37, "just to be sure there really is something"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_lazy_contour_basic():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    masks = ds["mask"][:]
    cont1 = get_contour_lazily(masks)
    cont2 = get_contour(masks)
    for ii in range(len(ds)):
        assert np.all(cont1[ii] == cont2[ii])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("idxs",
                         [slice(0, 5, 2),
                          [0, 2, 4],
                          np.array([0, 2, 4]),
                          np.array([True, False, True, False, True,
                                    False, False, False])
                          ])
def test_lazy_contour_slicing(idxs):
    h5path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    contours_ref = []
    with h5py.File(h5path, "a") as h5:
        indices = np.arange(len(h5["events"]["deform"]))[idxs]
        assert np.all(indices == [0, 2, 4])
        for idn in indices:
            contours_ref.append(h5["events"]["contour"][str(idn)][:])
        del h5["events"]["contour"]
        masks_ref = np.array(h5["events"]["mask"][:][idxs], dtype=bool)

    with new_dataset(h5path) as ds:
        # sanity check
        assert "contour" not in ds.features_innate
        contours_test = ds["contour"][idxs]

    for csoll, cist, mask in zip(contours_ref, contours_test, masks_ref):
        # we cannot compare the contours directly, because they may
        # have different starting points. So we compare the masks.
        mask_ist = np.zeros_like(mask)
        mask_ist[cist[:, 1], cist[:, 0]] = True
        mask_ist = binary_fill_holes(mask_ist)

        assert np.all(mask_ist == mask)
        # remove them so we don't make any mistakes in defining
        # the next mask.
        del cist
        del mask_ist

        mask_soll = np.zeros_like(mask)
        mask_soll[csoll[:, 1], csoll[:, 0]] = True
        mask_soll = binary_fill_holes(mask_soll)
        assert np.all(mask_soll == mask)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_lazy_contour_single():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    masks = ds["mask"][:]
    c_all = get_contour_lazily(masks)
    c_0 = get_contour_lazily(masks[0])
    c_1 = get_contour_lazily(masks[1])
    assert np.all(c_0 == c_all[0])
    assert np.all(c_1 == c_all[1])
    assert len(c_0) != len(c_all[1])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_lazy_contour_timing():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    masks = ds["mask"][:]
    t0 = time.perf_counter()
    get_contour_lazily(masks)
    t1 = time.perf_counter()
    get_contour(masks)
    t2 = time.perf_counter()
    assert t2-t1 > 10*(t1-t0)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_lazy_contour_type():
    ds1 = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    c1 = ds1["contour"]
    # force computation of contour data
    ds1._events._features.remove("contour")
    c2 = ds1["contour"]
    assert isinstance(c1, dclab.rtdc_dataset.fmt_hdf5.H5ContourEvent)
    assert isinstance(c2, dclab.features.contour.LazyContourList)


def test_simple_contour():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # Note: contour "3" in ds is bad!
    cin = ds["contour"][2]
    mask = np.zeros_like(ds["image"][2], dtype="bool")
    mask[cin[:, 1], cin[:, 0]] = True
    cout = get_contour(mask)
    # length
    assert len(cin) == len(cout)
    # simple presence test
    for ci in cin:
        assert ci in cout
    # order
    for ii in range(1, len(cin)):
        c2 = np.roll(cin, ii, axis=0)
        if np.all(c2 == cout):
            break
    else:
        assert False, "contours not matching, check orientation?"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_volume():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    mask = [mi for mi in ds["mask"]]
    cont1 = [ci for ci in ds["contour"]]
    cont2 = get_contour(mask)

    kw = dict(pos_x=ds["pos_x"],
              pos_y=ds["pos_y"],
              pix=ds.config["imaging"]["pixel size"])

    v1 = get_volume(cont=cont1, **kw)
    v2 = get_volume(cont=cont2, **kw)

    assert np.allclose(v1, v2)



"""Volume computation based on contour revolution"""
import numpy as np


def get_volume(cont, pos_x, pos_y, pix, fix_orientation=False):
    """Calculate the volume of a polygon revolved around an axis

    The volume estimation assumes rotational symmetry.

    Parameters
    ----------
    cont: ndarray or list of ndarrays of shape (N,2)
        A 2D array that holds the contour of an event [px]
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.
    pos_x: float or ndarray of length N
        The x coordinate(s) of the centroid of the event(s) [µm]
        e.g. obtained using `mm.pos_x`
    pos_y: float or ndarray of length N
        The y coordinate(s) of the centroid of the event(s) [µm]
        e.g. obtained using `mm.pos_y`
    pix: float
        The detector pixel size in µm.
        e.g. obtained using: `mm.config["imaging"]["pixel size"]`
    fix_orientation: bool
        If set to True, make sure that the orientation of the
        contour is counter-clockwise in the r-z plane
        (see :func:`vol_revolve`). This is False by default, because
        (1) Shape-In always stores the contours in the correct
        orientation and (2) there may be events with high porosity
        where "fixing" the orientation makes things worse and a
        negative volume is returned.

    Returns
    -------
    volume: float or ndarray
        volume in um^3

    Notes
    -----
    The computation of the volume is based on a full rotation of the
    upper and the lower halves of the contour from which the
    average is then used.

    The volume is computed radially from the the center position
    given by (`pos_x`, `pos_y`). For sufficiently smooth contours,
    such as densely sampled ellipses, the center position does not
    play an important role. For contours that are given on a coarse
    grid, as is the case for RT-DC, the center position must be
    given.

    References
    ----------
    - https://de.wikipedia.org/wiki/Kegelstumpf#Formeln
    - Yields identical results to the Matlab script by Geoff Olynyk
      <http://de.mathworks.com/matlabcentral/fileexchange/36525-volrevolve>`_
    """
    if np.isscalar(pos_x):
        cont = [cont]
        ret_list = False
    else:
        ret_list = True

    # Convert input to 1D arrays
    pos_x = np.atleast_1d(pos_x)
    pos_y = np.atleast_1d(pos_y)

    if pos_x.size != pos_y.size:
        raise ValueError("Size of `pos_x` and `pos_y` must match!")

    if pos_x.size > 1 and len(cont) <= 1:
        raise ValueError("Number of given contours too small!")

    # results are stored in a separate array initialized with nans
    v_avg = np.zeros_like(pos_x, dtype=float)*np.nan

    # v_avg has the shape of `pos_x`. We are iterating over the smallest
    # length for `cont` and `pos_x`.
    for ii in range(min(len(cont), pos_x.shape[0])):
        # If the contour has less than 4 pixels, the computation will fail.
        # In that case, the value np.nan is already assigned.
        cc = cont[ii]
        if cc.shape[0] >= 4:
            # Center contour coordinates with given centroid
            contour_x = cc[:, 0] - pos_x[ii] / pix
            contour_y = cc[:, 1] - pos_y[ii] / pix
            # Switch to r and z to follow notation of vol_revolve
            # (In RT-DC the axis of rotation is x, but for vol_revolve
            # we need the axis vertically)
            contour_r = contour_y
            contour_z = contour_x
            if fix_orientation:
                # Make sure the contour is counter-clockwise
                contour_r, contour_z = counter_clockwise(contour_r, contour_z)

            # Compute right volume
            # Which points are at negative r-values (r<0)?
            inx_neg = np.where(contour_r < 0)
            # These points will be shifted up to r=0 directly on the z-axis
            contour_right = np.copy(contour_r)
            contour_right[inx_neg] = 0
            vol_right = vol_revolve(contour_right, contour_x, pix)

            # Compute left volume
            # Which points are at positive r-values? (r>0)?
            idx_pos = np.where(contour_r > 0)
            # These points will be shifted down to y=0 to build an x-axis
            contour_left = np.copy(contour_r)
            contour_left[idx_pos] = 0
            # Now we still have negative r values, but vol_revolve needs
            # positive values, so we flip the sign...
            contour_left[:] *= -1
            # ... but in doing so, we have switched to clockwise rotation
            # and we need to pass the array in reverse order
            vol_left = vol_revolve(contour_left[::-1], contour_x[::-1], pix)

            # Compute the average
            v_avg[ii] = (vol_right + vol_left) / 2

    if not ret_list:
        # Do not return a list if the input contour was not in a list
        v_avg = v_avg[0]

    return v_avg


def counter_clockwise(cx, cy):
    """Put contour coordinates into counter-clockwise order

    Parameters
    ----------
    cx, cy: 1d ndarrays
        The x- and y-coordinates of the contour

    Returns
    -------
    cx_cc, cy_cc:
        The x- and y-coordinates of the contour in
        counter-clockwise orientation.

    Notes
    -----
    The contour must be centered around (0, 0).
    """
    # test orientation
    angles = np.unwrap(np.arctan2(cy, cx))
    grad = np.diff(angles)
    if np.average(grad) < 0:
        return cx[::-1], cy[::-1]
    else:
        return cx, cy


def vol_revolve(r, z, point_scale=1.):
    r"""Calculate the volume of a polygon revolved around the Z-axis

    This implementation yields the same results as the volRevolve
    Matlab function by Geoff Olynyk (from 2012-05-03)
    https://de.mathworks.com/matlabcentral/fileexchange/36525-volrevolve.

    The difference here is that the volume is computed using (a much
    more approachable) implementation using the volume of a truncated
    cone (https://de.wikipedia.org/wiki/Kegelstumpf).

    .. math::

      V = \frac{h \cdot \pi}{3} \cdot (R^2 + R \cdot r + r^2)

    Where :math:`h` is the height of the cone and :math:`r` and
    `R` are the smaller and larger radii of the truncated cone.

    Each line segment of the contour resembles one truncated cone. If
    the z-step is positive (counter-clockwise contour), then the
    truncated cone volume is added to the total volume. If the z-step
    is negative (e.g. inclusion), then the truncated cone volume is
    removed from the total volume.

    .. versionchanged:: 0.37.0

      The volume in previous versions was overestimated by on average
      2µm³.

    Parameters
    ----------
    r: 1d np.ndarray
        radial coordinates (perpendicular to the z axis)
    z: 1d np.ndarray
        coordinate along the axis of rotation
    point_scale: float
        point size in your preferred units; The volume is multiplied
        by a factor of `point_scale**3`.

    Notes
    -----
    The coordinates must be given in counter-clockwise order,
    otherwise the volume will be negative.
    """
    r = np.array(r).flatten()
    z = np.array(z).flatten()

    # sanity checks
    assert len(r) == len(z)
    assert len(r) >= 3
    assert len(r.shape) == len(z.shape) == 1
    assert np.all(r >= 0)

    # make sure we have a closed contour
    if (r[-1] != r[0]) or (z[-1] != z[0]):
        # We have an open contour - close it.
        r = np.resize(r, len(r) + 1)
        z = np.resize(z, len(z) + 1)

    rp = r[:-1]

    # array of radii differences: R - r
    dr = np.diff(r)
    # array of height differences: h
    dz = np.diff(z)

    # If we expand the function in the doc string with
    # dr = R - r and dz = h, then we get three terms for the volume
    # (as opposed to four terms in Olynyk's script). Those three terms
    # all resemble area slices multiplied by the z-distance dz.
    a1 = 3 * rp**2
    a2 = 3 * rp*dr
    a3 = dr**2

    # Note that the formula for computing the volume is symmetric
    # with respect to r and R. This means that it does not matter
    # which sign dr has (R and r are always positive). Since this
    # algorithm assumes that the contour is ordered counter-clockwise,
    # positive dz means adding to the contour while negative dz means
    # subtracting from the contour (see test functions for more details).
    # Conveniently so, dz only appears one time in this formula, so
    # we can take the sign of dz as it is (Otherwise, we would have
    # to take the absolute value of every truncated cone volume and
    # multiply it by np.sign(dz)).
    v = np.pi / 3 * dz * np.abs(a1 + a2 + a3)
    vol = np.sum(v) * point_scale ** 3

    return vol



"""Working with plugin features

This example shows how to load a user-defined plugin feature recipe in dclab
and use it in a scatter plot.

Please also download the :download:`plugin_example.py
<../examples/plugin_example.py>` file for this example.
"""
import pathlib

import matplotlib.pyplot as plt

import dclab


plugin_path = pathlib.Path(__file__).parent

# load a single plugin feature
dclab.load_plugin_feature(plugin_path / "plugin_example.py")

# load some data from DCOR
ds = dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0")

# access the features
circ_per_area = ds["circ_per_area"]
circ_times_area = ds["circ_times_area"]

# create a plot with a plugin feature
plt.figure(figsize=(8, 4))
xlabel = dclab.dfn.get_feature_label("circ_times_area")
ylabel = dclab.dfn.get_feature_label("deform")

ax1 = plt.subplot(title="Plot with a plugin feature")
ax1.plot(ds["circ_times_area"], ds["deform"],
         "o", color="k", alpha=.2, ms=1)
ax1.set_xlabel(xlabel)
ax1.set_ylabel(ylabel)
ax1.set_xlim(20, 40)
ax1.set_ylim(0.0025, 0.025)

plt.tight_layout()
plt.show()



"""Test machine learning tools"""
import pathlib
import tempfile

import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.rtdc_dataset import feat_anc_ml

from helper_methods import example_data_dict


data_path = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def cleanup_plugin_features():
    """Fixture used to cleanup plugin feature tests"""
    # code run before the test
    pass
    # then the test is run
    yield
    # code run after the test
    # remove our test plugin examples
    feat_anc_ml.remove_all_ml_features()


def make_data(add_feats=None, sizes=None):
    if sizes is None:
        sizes = [100, 130]
    if add_feats is None:
        add_feats = ["area_um", "deform"]
    keys = add_feats + ["time", "frame", "fl3_width"]
    data = []
    for size in sizes:
        data.append(new_dataset(example_data_dict(size=size, keys=keys)))
    return data


def test_af_ml_class_basic():
    data = {"ml_score_001": [.1, .3, .1, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    assert np.allclose(ds["ml_class"], [1, 0, 1, 0, 1])
    assert issubclass(ds["ml_class"].dtype.type, np.integer)


def test_af_ml_class_bad_feature():
    data = {"ml_score_0-1": [.1, .3, .1, 0.01, .59],
            }
    with pytest.raises(ValueError,
                       match="Invalid feature name 'ml_score_0-1'"):
        dclab.new_dataset(data)


def test_af_ml_class_bad_score_max():
    data = {"ml_score_001": [.1, .3, 99, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    with pytest.raises(ValueError, match="> 1"):
        ds["ml_class"]


def test_af_ml_class_bad_score_min():
    data = {"ml_score_001": [.1, .3, -.1, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    with pytest.raises(ValueError, match="< 0"):
        ds["ml_class"]


def test_af_ml_class_bad_score_nan():
    data = {"ml_score_001": [.1, .3, np.nan, np.nan, np.nan],
            "ml_score_002": [.2, .1, .4, np.nan, 0],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    assert np.allclose(ds["ml_class"],
                       [1, 0, 1, -1, -1],
                       equal_nan=True,
                       atol=0,
                       rtol=0
                       )


def test_af_ml_class_changed_features():
    data = {"ml_score_011": [.1, .3, .1, 0.01, .59],
            "ml_score_012": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    assert np.all(ds["ml_class"] == [1, 0, 1, 0, 1])
    # This triggers a recomputation of the ml_class feature the
    # next time it is accessed:
    ds._events["ml_score_003"] = np.array([1, 1, 1, 1, 0])
    assert np.all(ds["ml_class"] == [0, 0, 0, 0, 2])


def test_af_ml_class_has_ml_score_false():
    data = {"deform": [.1, .3, .1, 0.01, .59],
            "area_um": [20, 10, 40, 100, 80],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" not in ds


def test_af_ml_class_single():
    data = {"ml_score_001": [.1, .3, .1, 0.01, .59],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    assert np.allclose(ds["ml_class"], 0)


def test_af_ml_score_label_fallback():
    """Test whether the correct label is returned"""
    label1 = dclab.dfn.get_feature_label("ml_score_low")
    label2 = dclab.dfn.get_feature_label("ml_score_hig")
    assert label1 == "ML score LOW"
    assert label2 == "ML score HIG"


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.feat_anc_ml.modc.'
                            + 'ModelFormatExportFailedWarning')
def test_modc_export_model_bad_model():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    with pytest.raises(ValueError,
                       match="Export failed for all model formats!"):
        feat_anc_ml.modc.export_model(path=tmpdir,
                                      model=object()
                                      )



"""Reading and writing trained machine learning models for dclab"""
import atexit
import hashlib
import json
import pathlib
import shutil
import tempfile
import time
import traceback as tb
import warnings

from .ml_model import BaseModel


class ModelFormatExportFailedWarning(UserWarning):
    pass


def export_model(path, model, enforce_formats=None):
    """Export an ML model to all possible formats

    The model must be exportable with at least one method
    listed by :func:`BaseModel.all_formats`.

    Parameters
    ----------
    path: str or pathlib.Path
        Directory where the model is stored to. For each supported
        model, a new subdirectory or file is created.
    model: An instance of an ML model, NOT dclab.cfeat_anc_ml.models.BaseModel
        Trained model instance
    enforce_formats: list of str
        Enforced file formats for export. If the export for one
        of these file formats fails, a ValueError is raised.
    """
    if enforce_formats is None:
        enforce_formats = []
    path = pathlib.Path(path)
    if not path.is_dir():
        raise ValueError(
            "Output `path` should be a directory: '{}'".format(path))
    if len(sorted(path.rglob("*"))) != 0:
        raise ValueError(
            "Model output directory should be empty: '{}'".format(path))

    supported_formats = BaseModel.all_formats()
    if enforce_formats:
        for e_fmt in enforce_formats:
            if e_fmt not in supported_formats:
                raise ValueError(
                    "Unsupported format '{}', expected ".format(e_fmt)
                    + "one of {}!".format(", ".join(supported_formats.keys())))
    exported_formats = {}
    for fmt in supported_formats:
        tmp = tempfile.mkdtemp(prefix="dclab_ml_{}".format(fmt))
        try:
            suffix = supported_formats[fmt]["suffix"]
            tmp_out = pathlib.Path(tmp) / (fmt + suffix)
            cls = supported_formats[fmt]["class"]
            cls.save_bare_model(tmp_out, model, save_format=fmt)
            # attempt to load the model to see if it worked
            cls.load_bare_model(tmp_out)
        except BaseException:
            warnings.warn("Could not export to '{}': {}".format(
                fmt, tb.format_exc(limit=1)),
                ModelFormatExportFailedWarning)
            if fmt in enforce_formats:
                raise
        else:
            pout = path / tmp_out.name
            pathlib.Path(tmp_out).rename(pout)
            exported_formats[fmt] = tmp_out.name
        finally:
            shutil.rmtree(tmp, ignore_errors=True)

    if not exported_formats:
        raise ValueError("Export failed for all model formats!")

    # Now compute the hash and get the destination filename
    m_hash = hash_path(path)
    m_dict = {"sha256": m_hash,
              "formats": exported_formats,
              "date": time.strftime("%Y-%m-%d %H:%M"),
              }
    return m_dict


def hash_path(path):
    """Create a SHA256 hash of a file or all files in a directory

    The files are sorted before hashing for reproducibility.
    """
    path = pathlib.Path(path)
    assert path.is_dir()
    hasher = hashlib.sha256()
    if path.is_dir():
        paths = sorted(path.rglob("*"))
    else:
        paths = [path]
    for pp in paths:
        hasher.update(pp.name.encode())
        if pp.is_dir():
            continue
        with pp.open("rb") as fd:
            while True:
                chunk = fd.read(65536)
                if chunk:
                    hasher.update(chunk)
                else:
                    break
    return hasher.hexdigest()


def load_modc(path, from_format=None):
    """Load models from a .modc file for inference

    Parameters
    ----------
    path: str or path-like
        Path to a .modc file
    from_format: str
        If set to None, the first available format in
        :func:`BaseModel.all_formats` is used. If set to
        a key in :func:`BaseModel.all_formats`, then this
        format will take precedence and an error will
        be raised if loading with this format fails.

    Returns
    -------
    model: list of dclab.rtdc_dataset.feat_anc_ml.ml_model.BaseModel
        Models that can be used for inference via `model.predict`
    """
    # unpack everything
    t_dir = pathlib.Path(tempfile.mkdtemp(prefix="modc_load_"))
    cleanup = atexit.register(lambda: shutil.rmtree(t_dir, ignore_errors=True))
    shutil.unpack_archive(path, t_dir, format="zip")

    # Get the metadata
    with (t_dir / "index.json").open("r") as fd:
        meta = json.load(fd)

    assert meta["model count"] == len(meta["models"])
    supported_formats = BaseModel.all_formats()
    dc_models = []
    for model_dict in meta["models"]:
        mpath = t_dir / model_dict["path"]

        formats = list(model_dict["formats"].keys())
        if from_format:
            formats = [from_format] + formats

        for fmt in formats:
            if fmt in supported_formats:
                cls = supported_formats[fmt]["class"]
                load = cls.load_bare_model
                try:
                    bare_model = load(mpath / model_dict["formats"][fmt])
                except BaseException:
                    if from_format and fmt == from_format:
                        # user requested this format explicitly
                        raise
                else:
                    # load `bare_model` into BaseModel
                    model = cls(bare_model=bare_model,
                                inputs=model_dict["input features"],
                                outputs=model_dict["output features"],
                                info=model_dict,
                                )
                    break
            elif from_format and fmt == from_format:
                raise ValueError("The format specified via `from_format` "
                                 + " '{}' is not supported!".format(fmt))
        else:
            raise ValueError("No compatible model file format found!")
        dc_models.append(model)

    # We are nice and do the cleanup before exit
    cleanup()
    atexit.unregister(cleanup)

    return dc_models


def save_modc(path, dc_models):
    """Save ML models to a .modc file

    Parameters
    ----------
    path: str, pathlib.Path
        Output .modc path
    dc_models: list of/or dclab.rtdc_dataset.feat_anc_ml.models.BaseModel
        Models to save

    Returns
    -------
    meta: dict
        Dictionary written to index.json in the .modc file
    """
    if not isinstance(dc_models, list):
        dc_models = [dc_models]

    # save all models to a temporary directory
    t_dir = pathlib.Path(tempfile.mkdtemp(prefix="modc_save_"))
    cleanup = atexit.register(lambda: shutil.rmtree(t_dir, ignore_errors=True))

    model_data = []
    for ii, mm in enumerate(dc_models):
        p_mod = t_dir / "model_{}".format(ii)
        p_mod.mkdir()
        m_dict = export_model(p_mod, mm.bare_model)
        m_dict["index"] = ii
        m_dict["input features"] = mm.inputs
        m_dict["output features"] = mm.outputs
        m_dict["path"] = p_mod.name
        for key in ["description", "long description", "output labels"]:
            if key in mm.info:
                m_dict[key] = mm.info[key]
        model_data.append(m_dict)

    meta = {
        "model count": len(dc_models),
        "models": model_data,
    }

    # save metadata
    with (t_dir / "index.json").open("w") as fd:
        json.dump(meta, fp=fd, indent=2, separators=(',', ': '),
                  sort_keys=True)

    aout = shutil.make_archive(base_name=path, format="zip", root_dir=t_dir,
                               base_dir=".")
    pathlib.Path(aout).rename(path)

    # We are nice and do the cleanup before exit
    cleanup()
    atexit.unregister(cleanup)

    return meta



from distutils.version import LooseVersion
import importlib
import os
import warnings

#: Minimum R version
#: This is actually a dependency for rpy2, because the API changed then
#: (ffi.error: symbol 'R_tryCatchError' not found in library).
R_MIN_VERSION = "3.6.0"

#: Minimum rpy2 version
RPY2_MIN_VERSION = "2.9.4"

R_SUBMODULES = [
    "rpy2.robjects.packages",
    "rpy2.situation",
    "rpy2.robjects.vectors",
]

R_SUBMODULES_3 = [
    "rpy2.rinterface_lib.callbacks",
]


class RPY2UnavailableError(BaseException):
    pass


class RPY2ImportError(RPY2UnavailableError):
    pass


class RPY2OutdatedError(RPY2UnavailableError):
    pass


class RUnavailableError(BaseException):
    pass


class ROutdatedError(RUnavailableError):
    pass


class MockRPackage:
    def __init__(self, exception):
        self.exception = exception

    def __getattr__(self, item):
        raise self.exception


def import_r_submodules():
    importlib.import_module("rpy2.situation")
    r_home = rpy2.situation.get_r_home()
    if r_home is not None:
        if os.environ.get("R_HOME", None) is None:
            # set R_HOME globally (https://github.com/rpy2/rpy2/issues/796)
            os.environ["R_HOME"] = r_home
        if rpy2_is_version_3:
            mods = R_SUBMODULES + R_SUBMODULES_3
        else:
            mods = R_SUBMODULES
        try:
            for sm in mods:
                importlib.import_module(sm)
        except rpy2.rinterface_lib.openrlib.ffi.error as exc:
            # This error happens when the installed R version is too old:
            # "ffi.error: symbol 'R_tryCatchError' not found in library"
            raise ROutdatedError(
                f"Encountered '{exc.__class__.__name__}: {exc}'. "
                f"Please make sure you have 'R>={R_MIN_VERSION}'!")


try:
    rpy2 = importlib.import_module("rpy2")
    if LooseVersion(rpy2.__version__) < LooseVersion(RPY2_MIN_VERSION):
        raise RPY2OutdatedError(f"Please install 'rpy2>={RPY2_MIN_VERSION}'!")
except ImportError:
    rpy2 = MockRPackage(
        RPY2ImportError(f"Please install 'rpy2>={RPY2_MIN_VERSION}'!"))
    rpy2_is_version_3 = False
except BaseException as e:
    rpy2 = MockRPackage(e)
    rpy2_is_version_3 = False
else:
    rpy2_is_version_3 = LooseVersion(rpy2.__version__) >= LooseVersion("3.0")
    try:
        import_r_submodules()
    except RUnavailableError as e:
        warnings.warn("There is an issue with the linked R version: "
                      + f"{e.__class__.__name__}: {e}")



"""Test hdf5 file format"""
import os

import h5py
import numpy as np
import pytest

import dclab
from dclab import new_dataset, rtdc_dataset
from dclab.rtdc_dataset import config

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_config():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert ds.config["setup"]["channel width"] == 30
    assert ds.config["setup"]["chip region"].lower() == "channel"
    assert ds.config["setup"]["flow rate"] == 0.16
    assert ds.config["imaging"]["pixel size"] == 0.34


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_contour_basic():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert len(ds) == 5
    assert len(ds["contour"]) == 5
    assert np.allclose(np.average(ds["contour"][0]), 30.75)
    assert np.median(ds["contour"][0]) == 30.5
    assert np.median(ds["contour"][2]) == 31.5
    assert np.median(ds["contour"][-3]) == 31.5
    assert np.median(ds["contour"][4]) == 33.0
    assert np.median(ds["contour"][-1]) == 33.0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_defective_feature_aspect():
    # see https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut/issues/241
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # modify aspect feature
    with h5py.File(h5path, "a") as h5:
        aspect0 = h5["events/aspect"][0]
        aspect1 = 1.234
        assert not np.allclose(aspect0, aspect1), "test's sanity"
        h5["events/aspect"][0] = aspect1
        # In Shape-In 2.0.5 everything worked fine
        h5.attrs["setup:software version"] = "ShapeIn 2.0.5"
    # sanity check
    with new_dataset(h5path) as ds1:
        assert np.allclose(ds1["aspect"][0], aspect1)
    # trigger recomputation of aspect feature
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.6"
    # verify original value of aspect
    with new_dataset(h5path) as ds2:
        assert np.allclose(ds2["aspect"][0], aspect0)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_defective_feature_volume():
    # see https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut/issues/241
    h5path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    with h5py.File(h5path, "r") as h5:
        # This volume was computed with an old version of dclab
        wrong_volume = h5["events/volume"][:]

    with new_dataset(h5path) as ds:
        assert "volume" not in ds.features_innate
        correct_volume = ds["volume"]
        assert not np.allclose(wrong_volume, correct_volume)

    # prevent recomputation via logs
    # (do not use context manager here [sic])
    hw = dclab.RTDCWriter(h5path)
    hw.store_log("dclab_issue_141", ["fixed"])
    hw.h5file.close()
    with new_dataset(h5path) as ds2:
        assert np.allclose(ds2["volume"], wrong_volume)

    # remove the log
    with h5py.File(h5path, "a") as h5:
        del h5["logs/dclab_issue_141"]

    # make sure that worked
    with new_dataset(h5path) as ds:
        assert "volume" not in ds.features_innate
        correct_volume = ds["volume"]
        assert not np.allclose(wrong_volume, correct_volume)

    # prevent recomputation via dclab version string
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.6 | dclab 0.37.0"
    with new_dataset(h5path) as ds2:
        assert "volume" in ds2.features_innate
        assert np.allclose(ds2["volume"], wrong_volume)

    # reset version string
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.6 | dclab 0.35.1"
    with new_dataset(h5path) as ds2:
        assert "volume" not in ds2.features_innate
        assert np.allclose(ds2["volume"], correct_volume)

    # use context manager to write version number
    with dclab.RTDCWriter(h5path) as hw:
        pass
    with new_dataset(h5path) as ds2:
        assert "volume" in ds2.features_innate
        assert np.allclose(ds2["volume"], wrong_volume)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_hash():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert ds.hash == "2c436daba22d2c7397b74d53d80f8931"
    assert ds.format == "hdf5"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_load_non_scalar_data():
    """Loading non-scalar data should return data as h5py.Dataset"""
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert isinstance(ds["image"], h5py.Dataset)
    assert isinstance(ds["image_bg"], h5py.Dataset)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_contour():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    assert "contour" in ds.features_innate
    assert len(ds["contour"]) == 8
    assert ds["contour"].shape == (8, np.nan, 2)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_image():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert "image" in ds.features_innate
    assert len(ds["image"]) == 5
    assert ds["image"].shape == (5, 80, 250)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_mask():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert "mask" in ds.features_innate
    assert len(ds["mask"]) == 5
    assert ds["mask"].shape == (5, 80, 250)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_trace():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert len(ds) == 7
    assert "trace" in ds.features_innate
    assert ds["trace"].shape == (6, 7, 177)
    assert ds["trace"]["fl1_raw"].shape == (7, 177)
    assert ds["trace"]["fl1_raw"][0].shape == (177,)
    assert len(ds["trace"]) == 6
    assert len(ds["trace"]["fl1_raw"]) == 7
    assert len(ds["trace"]["fl1_raw"][0]) == 177


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ignore_empty_hdf5_meta_data_attribute():
    """Ignore empty hdf5 attributes / dclab metadata"""
    # see https://github.com/ZELLMECHANIK-DRESDEN/dclab/issues/109
    path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # add empty attribute
    with h5py.File(path, "r+") as h5:
        h5.attrs["setup:module composition"] = ""

    # assert
    with pytest.warns(config.EmptyConfigurationKeyWarning,
                      match=r"Empty value for \[setup\]: 'module composition'!"
                      ):
        new_dataset(path)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ignore_unknown_hdf5_meta_data_attribute():
    """Ignore unknown hdf5 attributes / dclab metadata"""
    # see https://github.com/ZELLMECHANIK-DRESDEN/dclab/issues/109
    path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # add empty attribute
    with h5py.File(path, "r+") as h5:
        h5.attrs["setup:shizzle"] = ""

    # assert
    with pytest.warns(config.UnknownConfigurationKeyWarning,
                      match=r"Unknown key 'shizzle' in the 'setup' section!"
                      ):
        new_dataset(path)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_image_basic():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert np.allclose(np.average(ds["image"][1]), 125.37133333333334)
    assert len(ds["image"]) == 5


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_image_bg():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    # add a fake image_bg column
    with rtdc_dataset.RTDCWriter(path) as hw:
        image_bg = hw.h5file["events"]["image"][:] // 2
        hw.store_feature("image_bg", image_bg)

    with new_dataset(path) as ds:
        for ii in range(len(ds)):
            assert np.all(ds["image"][ii] // 2 == ds["image_bg"][ii])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_image_bg_2():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with new_dataset(path) as ds:
        assert "image_bg" in ds
        bgc = ds["image"][0] - ds["image_bg"][0]
        assert bgc[10, 11] == 6


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
@pytest.mark.parametrize("idxs",
                         [slice(0, 5, 2),
                          [0, 2, 4],
                          np.array([0, 2, 4]),
                          np.array([True, False, True, False, True])])
def test_index_slicing_contour(idxs):
    data = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(data)

    contour_ref = [
        ds["contour"][0],
        ds["contour"][2],
        ds["contour"][4],
    ]

    contour_slice = ds["contour"][idxs]

    assert np.all(contour_slice[0] == contour_ref[0])
    assert np.all(contour_slice[1] == contour_ref[1])
    assert np.all(contour_slice[2] == contour_ref[2])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    with new_dataset(path_in) as ds:
        assert not ds.logs

    # write some logs
    with h5py.File(path_in, "a") as h5:
        hw = rtdc_dataset.RTDCWriter(h5)
        hw.store_log("test_log",  ["peter", "hans"])

    with new_dataset(path_in) as ds:
        assert ds.logs
        assert ds.logs["test_log"][0] == "peter"

    # remove logs
    with h5py.File(path_in, "a") as h5:
        del h5["logs"]

    with new_dataset(path_in) as ds:
        assert not ds.logs
        try:
            ds.logs["test_log"]
        except KeyError:  # no log data
            pass


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_no_suffix():
    """Loading an .rtdc file that has a wrong suffix"""
    path = str(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    path2 = path + ".wrong_suffix"
    os.rename(path, path2)
    ds = new_dataset(path2)
    assert(len(ds) == 8)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_online_polygon_filters():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # add an artificial online polygon filter
    with h5py.File(path, "a") as h5:
        # set soft filter to True
        h5.attrs["online_filter:area_um,deform soft limit"] = True
        # set filter values
        pf_name = "online_filter:area_um,deform polygon points"
        area_um = h5["events"]["area_um"]
        deform = h5["events"]["deform"]
        pf_points = np.array([
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform)],
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform) + np.std(deform)],
            [np.mean(area_um),
             np.mean(deform) + np.std(deform)],
        ])
        h5.attrs[pf_name] = pf_points

    # see if we can open the file without any error
    with new_dataset(path) as ds:
        assert len(ds) == 8
        assert ds.config["online_filter"]["area_um,deform soft limit"]
        assert "area_um,deform polygon points" in ds.config["online_filter"]
        assert np.allclose(
            ds.config["online_filter"]["area_um,deform polygon points"],
            pf_points)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_open_with_invalid_feature_names():
    """Loading an .rtdc file that has wrong feature name"""
    path = str(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    # add wrong feature name right at the top of the list
    with h5py.File(path, "r+") as h5:
        h5["events"]["a0"] = h5["events"]["deform"]
    # see if we can open the file without any error
    ds = new_dataset(path)
    assert len(ds) == 8


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_trace():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert len(ds["trace"]) == 2
    assert ds["trace"]["fl1_raw"].shape == (5, 100)
    assert np.allclose(np.average(
        ds["trace"]["fl1_median"][0]), 0.027744706519425219)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in sorted(list(loc.keys())):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import h5py
import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.features.bright import get_bright

from helper_methods import calltracker, retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with h5py.File(path, "r+") as h5:
        real_avg = h5["events"]["bright_avg"][:]
        real_sd = h5["events"]["bright_sd"][:]
        del h5["events"]["bright_avg"]
        del h5["events"]["bright_sd"]
    ds = dclab.new_dataset(path)
    # sanity checks
    assert "bright_avg" not in ds.features_innate
    assert "bright_sd" not in ds.features_innate
    comp_avg = ds["bright_avg"]
    comp_sd = ds["bright_sd"]
    idcompare = ~np.isnan(comp_avg)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_avg[idcompare], comp_avg[idcompare])
    assert np.allclose(real_sd[idcompare], comp_sd[idcompare])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_brightness_called_once(monkeypatch):
    """Make sure dclab.features.bright.get_bright is only called once"""
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # remove brightness features
    with h5py.File(path, "r+") as h5:
        del h5["events"]["bright_avg"]
        del h5["events"]["bright_sd"]

    # wrap the original brightness retrieval function
    old_get_bright = dclab.features.bright.get_bright

    @calltracker
    def wrap_get_bright(*args, **kwargs):
        return old_get_bright(*args, **kwargs)
    # Monkeypatch the imported function used in the feat_anc_core submodule,
    # otherwise (dclab.features.bright.get_bright) it is not called.
    monkeypatch.setattr("dclab.rtdc_dataset.feat_anc_core.af_image_contour"
                        ".features.bright.get_bright",
                        wrap_get_bright)

    # assert (access both brightness features and make sure that the
    # original function is only called once)
    ds = dclab.new_dataset(path)
    _ = ds["bright_sd"]
    _ = ds["bright_avg"]

    assert wrap_get_bright.calls == 1


def test_simple_bright():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    for ii in range(2, 7):
        # This stripped dataset has only 7 video frames / contours
        image = ds["image"][ii]
        mask = ds["mask"][ii]
        avg, std = get_bright(mask=mask, image=image, ret_data="avg,sd")
        assert np.allclose(avg, ds["bright_avg"][ii])
        assert np.allclose(std, ds["bright_sd"][ii])
        # cover single `ret_data` input
        assert np.allclose(
            avg, get_bright(mask=mask, image=image, ret_data="avg"))
        assert np.allclose(
            std, get_bright(mask=mask, image=image, ret_data="sd"))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
